# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""
import math
import datetime

import matplotlib.pyplot as plt
import matplotlib
import pandas as pd
import numpy as np

from hsr1.plots import graphUtils

class DailyHists:
    def __init__(self, data, fig=None):
        data["pc_time_end_measurement"] = data["pc_time_end_measurement"].dt.tz_localize(None)
        dates = data["pc_time_end_measurement"].dt.date
        np_dates = dates.to_numpy()
        np_dates.sort()
        dates = pd.Series(np_dates)
        data["pc_time_end_measurement"] = pd.DatetimeIndex(dates)
        
        self.dates = dates
        self.data = data
        
        self.fig = fig
        
        ##### weight each day to the number of readings in the day
        weights = []
        for date in dates.unique():
            ##### TODO: \/\/ slow!
            readings = dates.loc[dates==date]
            weights += [8640/len(readings)]*(len(readings))
        
        self.weights = np.array(weights)
        
        
        self.first_date = str(dates[0])
        self.last_date = str(dates[len(dates)-1])
        self.date_range = len(pd.date_range(self.first_date, self.last_date, freq="D"))
    
    
    def plot_hists(self, columns, title, ignore_zero=False, weight=False, **kwargs):
        """plots several hists on one page
        params:
            columns: list of columns names to plot, OR string of one column name
            title: title of the plot
            ignore_zero: filters out all the zero values
            weight: weights each column of the histogram based on how many readings are in that day
                    means shorter days have the same impact as longer days        
        """
        fig, axes = plt.subplots(len(columns), figsize=(16.5, 11.7))
        fig.suptitle(title)
        fig.tight_layout()
        
        if len(columns) == 1:
            self.plot_one_hist(columns[0], axes, ignore_zero, weight=weight, **kwargs)
        else:
            for i, column in enumerate(columns):
                self.plot_one_hist(column, axes[i], ignore_zero, weight=weight, **kwargs)
    
    def plot_one_hist(self, column, axes, 
                      ignore_zero:bool=False,
                      weight:bool=True,
                      bins=None,  
                      zero_axes:bool=False, 
                      ylims:(int, int)=None, 
                      show_xticks:bool=True, 
                      log:bool=False,
                      ybuffer:bool=False,
                      limited_bins:bool=False,
                      linthresh:int=10,
                      **kwargs):
        """plots a single histogram on the given axes
        params:
            column: the column in the dataframe that is being plotted
            axes: the axes that the histogram is being plotted on
            ignore_zero: filters out all the zero values
            weight: weights each column of the histogram based on how many readings are in that day
                    means shorter days have the same impact as longer days
            bins: number of bins to have on the y-axis
            zero_axes: if True, starts the y-axis from 0, rather than the lowest data value
            ylims: can manually set the ylims
            show_xticks: wether or not to show labels on the x-axis, usefull when plotting multiple at once with the same scale
            log: if true, uses a log scale in the positive and negative direction
            ybuffer:  if true, adds a 10% buffer to the top of the graph
            limited_bins: if true, sets the number of bins to the number of unique readings
            
            linthresh: when using a log scale, how far on either side of zero to have a linear scale rather than log
        """
        data = self.data.copy()
        
        # data["pc_time_end_measurement"] = data["pc_time_end_measurement"].dt.tz_localize(None)
        
        weights = self.weights
        not_nan = data[column].isna()
        data = data.loc[~not_nan]
        weights = weights[~not_nan]
        
        column_data = data[column].to_numpy()
        
        ##### if all the data is the same value
        one_value = len(np.unique(column_data)) == 1
        
        ybin = bins
        non_zero_df = data[column]#.replace(0, np.nan)
        if ybin is None:
            if not non_zero_df.isnull().all():
                ##### min bin is 10, max is 100
                
                space_covered = max(10, int((np.nanmax(non_zero_df)-np.nanmin(non_zero_df))*1000))
                ybin = min(100, space_covered)
            else:
                ybin = 11
            
            if limited_bins:
                data_range = np.max(non_zero_df) - np.min(non_zero_df)
                if zero_axes:
                    data_range = np.max(non_zero_df)+1
                min_diff = np.min(np.diff(np.sort(np.unique(non_zero_df))))
                
                ybin = int(data_range/min_diff)
                 
            if one_value:
                ybin = 10
        
        
        
        bins = [self.date_range, ybin]
        
        if log:
            max_mag = np.max(np.abs(column_data))
            if max_mag > linthresh:
                column_data, ylims = self.scale(axes, column_data, ylims, linthresh, **kwargs)
        
        
        
        
        
        start = datetime.datetime(1970,1,1)
        a_range = [[(data["pc_time_end_measurement"].iloc[0]-start).total_seconds()*10e8, (data["pc_time_end_measurement"].iloc[-1]-start).total_seconds()*10e8],
                  [min(column_data), max(column_data)]]
        
        if ignore_zero and ((data[column] > 0).all() or (data[column] < 0).all()):
            non_zero = ((data[column] > 0) | (data[column] < 0)).to_numpy()
            data = data.loc[non_zero]
            column_data = column_data[non_zero]
            weights = weights[non_zero]
            a_range[1][0] = min(column_data) if len(column_data) != 0 else 0
        
        if ylims is None:
            if zero_axes:
                a_range[1][0] = 0
        if ylims is not None:
            a_range[1] = [ylims[0], ylims[1]]
        
        hist = None
        if len(data) != 0:
            # if weight:
            #     hist = axes.hist2d(data["pc_time_end_measurement"], column_data, bins=bins, cmap="jet", cmin=1, range=a_range, weights=weights)
            # else:
            #     hist = axes.hist2d(data["pc_time_end_measurement"], column_data, bins=bins, cmap="jet", cmin=1, range=a_range)
            
            if weight:
                hist_array, bin_edges_x, bin_edges_y = np.histogram2d(data["pc_time_end_measurement"], column_data, bins=bins, range=a_range, weights=weights)
            else:
                hist_array, *_ = np.histogram2d(data["pc_time_end_measurement"], column_data, bins=bins, range=a_range)
            
            hist_array = hist_array.T
            hist_array = np.flip(hist_array, 0)
            hist_array[hist_array == 0.0] = np.nan
        
            extent=[a_range[0][0], a_range[0][1], a_range[1][0], a_range[1][1]]
            
            if extent[2] == extent[3]:
                extent[3] += 1
            
            if extent[0] == extent[1]:
                extent[1] += 60*1440*10e8
            
            
            hist = axes.imshow(hist_array, cmap="jet", aspect="auto", extent=extent, interpolation="none")
            
        
            xticks, labels = graphUtils.calculate_date_labels(pd.Series(pd.date_range(self.first_date, self.last_date, freq="D")), axes.get_xlim())
            if show_xticks:
                axes.set_xticks(xticks, labels)
            else:
                axes.set_xticks(xticks, [])
        
        
        
        if one_value:
            axes.set_yticks([column_data[0]], [column_data[0]])
        
        
        axes.set_ylabel(column.lstrip("_"))
        if zero_axes:
            if ylims is None:
                ylims = (0, np.max(column_data))
            else:
                ylims = (0, ylims[1])
        
        if ybuffer:
            if ylims is None:
                ylims = axes.get_ylim()
            ylims = ylims[0], ylims[1]*1.1
        
        if ylims is not None:
            axes.set_ylim(ylims[0], ylims[1])
        
        if self.fig is not None:
            self.fig.colorbar(hist, ax=axes, format="%.0f")
        
        
        
    
    def scale(self, axes, 
              column_data, 
              ylim=None, 
              linthresh=10, 
              base=10, 
              linscale=2, 
              lin_ticks=5, 
              log_ticks_skipped=0):
        """applies a symmetrical log scale to the data, and also the axes, before it gets binned"""
        sym_log_scale = matplotlib.scale.SymmetricalLogScale(None, base=base, linthresh=linthresh, linscale=linscale)
        sym_log_transform = sym_log_scale.get_transform()
        
        ##### find the max base to know how many ticks to make
        abs_max = np.max(np.abs(column_data))
        tick_max = np.emath.logn(base, abs_max)
        ##### round up to the nearest int
        tick_max = int(tick_max+1)
        
        ##### apply log trasnformation to data
        column_data = sym_log_transform.transform_non_affine(column_data)
        
        ##### generate ticks for log part of graph
        sample_yticks = np.array([base**i for i in range(tick_max+1)])
        sample_log_yticks = sample_yticks[sample_yticks > linthresh]
        sample_log_yticks_selected = sample_log_yticks[::log_ticks_skipped+1]
        
        ##### generate ticks for linear part of graph
        sample_lin_yticks = np.linspace(0, linthresh, lin_ticks).astype(int)
        
        ##### merge lin and log
        sample_yticks = np.concatenate((sample_lin_yticks, sample_log_yticks))
        
        
        ##### apply log transformation to log part of the scale and recombine with linear part
        logged_sample_log_yticks = sym_log_transform.transform_non_affine(sample_yticks)
        logged_sample_yticks = logged_sample_log_yticks
        
        ##### flip and invert for negative part
        logged_sample_yticks = np.concatenate((-np.flip(logged_sample_yticks), logged_sample_yticks))
        sample_yticks = np.concatenate((-np.flip(sample_yticks), sample_yticks))
        
        ##### format ylabels
        labels = []
        for i, tick in enumerate(sample_yticks):
            ##### if tick is part of log section
            if (tick > linthresh or tick < -linthresh):
                if tick in sample_log_yticks_selected or -tick in sample_log_yticks_selected:
                    _tick = tick
                    prefix = ""
                    if tick <= -linthresh:
                        prefix = "-"
                        _tick = -tick
                    power = math.log(_tick, base)
                    labels.append(prefix+str(base)+graphUtils.superscript(int(round(power))))
                else:
                    labels.append("")
            else:
                labels.append(str(tick))
        
        if ylim is None:
            ylim = (min(column_data), max(column_data))
            ylim = (min(0, ylim[0]), max(0, ylim[1]))
            ylim = np.array(ylim)
        else:
            ylim = sym_log_transform.transform_non_affine(np.array(ylim))
        
        
        ##### apply calculated yticks using their before log values as labels  
        axes.set_yticks(logged_sample_yticks, labels)
        
        return column_data, ylim
    
    
    


# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License (LGPL)
"""

from hsr1.dbDriver import DBDriver
from hsr1.graph import Graph

import hsr1.synthetic_dataset
import hsr1.read_txt


# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""
import os
import time

import pandas as pd
import numpy as np
import sqlite3

from hsr1.db import Serialisation
from hsr1.utils import ReformatData as reformat

class SqliteDBLoad():
    """Handles all the loading from the database"""
    def __init__(self, db_name:str):
        """
        params:
            db_name: the name of the db you are loading from
        """
        if os.path.isfile(db_name):
            self.exists = True
        else:
            self.exists = False
        self.db_name = db_name
    
    
    
    
    def load_table_names(self) -> dict:
        """return a dictionary of all the table names and their corresponding column names"""
        con = sqlite3.connect(self.db_name)
        cur = con.cursor()
        
        ##### get all the table names
        sql = "SELECT name FROM sqlite_schema WHERE type='table'"
        cur.execute(sql)
        table_names = [table[0] for table in cur.fetchall()]
        
        ##### get all the column headers for each table
        column_headers = {}
        for table in table_names:
            sql = "PRAGMA table_info("+table+");"
            cur.execute(sql)
            column_headers[table] = [column[1] for column in cur.fetchall()]
        
        con.close()
        return column_headers
    
    
    def load(self, columns:[str]=[], 
             table:str=None,
             start_time:str=None, 
             end_time:str=None,
             condition:str="",
             raise_on_missing:bool=True,
             sort:bool=True,
             deserialise:bool=True,
             timezone:str="+00:00",
             **kwargs,
             ) -> pd.DataFrame:
        ##### TODO: fix docstring
        """loads data from the database
        params:
            columns: list of column names to extract from database, if empty, returns all
            table: which table to load data from. Not necessary but can be useful to specify when multiple tables have the same column
            start_time, end_time: for only selecting data within a time period
                format: ISO-8601("2023-05-23 00:00:00")
                if you dont put in a full datetime, it will assume first timestamp,
                    e.g: 2023-05-23 will be read as 2023-05-23 00:00:00.
                    therefore, start_time is inclusive, end time is not
            condition: sql that comes after the WHERE in an sql statement.
                e.g: gobal_integral>50
                if you want even more control, use load_sql
            raise_on_missing: how to handle when columns arent found, raise or just load if possible
            sort: sorts the results from the database
            deserialise: wether or not to automatically deserialise spectral data
            timezone: the timezone that all the data will be converted to
        
        returns:
            dataframe with the requested values. deployment metadata values are just at the start of the dataframe,
            with null values for the rest of the column
        
        if the start and end times dont include the dataset, it will return an empty dataframe,
        if the start or end times are invalid, it will ignore it and include the whole dataset, subject to other conditions
        """
        if not os.path.isfile(self.db_name):
            raise Exception(f"database '{self.db_name}' does not exist")
        column_headers = self.load_table_names()
        all_column_headers = self.load_table_names()
        if "raw_data" in column_headers.keys():
            del column_headers["raw_data"]
        
        
        ##### if table is specified, check all columns exist
        if table is not None:
            for column in columns:
                if not column in column_headers[table] and raise_on_missing:
                    raise KeyError("column: '" +column+ "' was not found in table "+table)
        
        ##### remove duplicates
        non_dupe_columns = []
        for col in columns:
            if col not in non_dupe_columns:
                non_dupe_columns.append(col)
        columns = non_dupe_columns
        
        ##### make a list of all the column headers except accessory and a list of all the accessory column headers
        if table == "accessory":
            table_type = "accessory"
        else:
            column_headers, table_type = self.__guess_type(column_headers, columns)
        
        ##### find all table names and all column names for each table
        ##### nested for loops, but it will never have high iterations, executes in 1x10^-5s
        table_and_column_names = []
        output_columns = []
        for column in columns:
            found = False
            if table_type == "accessory":
                table_and_column_names.append(column)
                output_columns.append(column)
                found = True
            else:
                for key in column_headers:
                    if column in column_headers[key] and key != "raw":
                        found = True
                        table_and_column_names.append(key+"."+column)
                        output_columns.append(column)
                        break
            if not found and raise_on_missing:
                raise KeyError("column: '" +column+ "' was not found, maybe use 'load_metadata()' or 'load_accessory_data()")
        
        
        
        

        if columns == []:
            table_and_column_names = ["*"]
            if table is None:
                output_columns = all_column_headers["spectral_data"]+all_column_headers["system_data"]+all_column_headers["precalculated_values"]
            else:
                output_columns = all_column_headers[table]
        
        
        time_col = ""
        if table is not None:
            time_col = table+".pc_time_end_measurement"
        elif table_type == "normal":
            time_col = "spectral_data.pc_time_end_measurement"
        elif table_type == "accessory":
            time_col = "accessory_data.pc_time_end_measurement"
        
        
        
        time_condition = " WHERE "
        if start_time != None:
            if table_type == "deployment":
                raise Exception("selecting by time is invalid for deployment_metadata")
            times_after = time_col+" > \""+str(start_time)+"\" and "
            time_condition += times_after
        
        if end_time != None:
            if table_type == "deployment":
                raise Exception("selecting by time is invalid for deployment_metadata")
            times_before =  time_col+" < \""+str(end_time)+"\" and "
            time_condition += times_before
        
        if len(condition)>0:
            condition = time_condition + condition
        ##### check if there is a time condition(7 characters is " WHERE ")
        elif len(time_condition)>7:
            condition = time_condition[:-4]
        
        result = pd.DataFrame()
        if table is None:
            ##### fetch data from spectral data table and system data
            if len(table_and_column_names)>0 or columns == []:
                
                tables = list(column_headers.keys())
                
                main_table = tables[0]
                table_string = main_table
                for table_name in tables[1:]:
                    table_string += " LEFT JOIN " + table_name + " ON " + main_table+".sample_id = " + table_name + ".sample_id"
                sql = ""
                
                order_by = ""
                if "pc_time_end_measurement" in columns:
                    order_by = " ORDER BY STRFTIME('%s', "+time_col+")"
                
                if table_type == "accessory":
                    sql = "SELECT "+", ".join(table_and_column_names)+" FROM accessory_data"+condition
                    if sort:
                        sql += order_by
                elif table_type == "deployment":
                    sql = "SELECT "+", ".join(table_and_column_names)+" FROM deployment_metadata"+condition
                else:
                    sql = "SELECT "+", ".join(table_and_column_names)+" FROM " + table_string +condition
                    if sort:
                        sql += order_by
                
                result = self.load_sql(sql)
                
                if result is None:
                    raise ValueError("no data was returned by the query: "+sql)
                
                if not result.empty:
                    result.columns = output_columns
        
        else:
            ##### table has been specified
            column_list = ", ".join(columns)
            if columns == []:
                column_list = "*"
            sql = "SELECT " + column_list + " FROM "+table+condition
            result = self.load_sql(sql)
            if not result.empty:
                result.columns = output_columns
        
        ##### remove duplicated columns
        result = result.loc[:, ~result.columns.duplicated()].copy()
        
        if deserialise:
            ser = Serialisation("numpy")
            if "global_spectrum" in output_columns:
                result = ser.decode_dataframe(result, ["global_spectrum"])
            if "diffuse_spectrum" in output_columns:
                result = ser.decode_dataframe(result, ["diffuse_spectrum"])
        
        
        if "pc_time_end_measurement" in result.columns:
            timezone, timedelta = reformat().calculate_timezone(timezone)
            result["pc_time_end_measurement"] = pd.to_datetime(result["pc_time_end_measurement"], utc=True)
            result["pc_time_end_measurement"] = result["pc_time_end_measurement"].dt.tz_convert(timezone)#.dt.tz_localize(None)
        
        if len(result) == 0:
            raise ValueError("No data matches your query, check your start_time, end_time and any conditions")
        
        return result
    
    
    def load_raw(self, columns:[str]=[], 
                 start_time=None, 
                 end_time=None):
        """loads the raw data from the raw_data table
        
        params: 
            columns: list of columns that you want to load, they are in the format:
                "channel_0", "channel_1" etc.
                if [], all columns are loaded
            start_time, end_time: for only selecting data within a time period
                format: ISO-8601("2023-05-23 00:00:00")
                if you dont put in a full datetime, it will assume first timestamp,
                    e.g: 2023-05-23 will be read as 2023-05-23 00:00:00.
                    therefore, start_time is inclusive, end time is not
        returns:
            a dataframe with the requested data. pc_time_end_measurement is automatically added as a column
            columns = pc_time_end_measurement, requested_channel_0, requested_channel_1
            if only one column is requested, it is returned as one unpacked dataframe, index=time, columns=wavelength
        """
        parameter_columns = columns
        if columns == []:
            columns = self.load_table_names()
            
            if not "raw_data" in columns:
                raise ValueError("this database dosent have raw_data")
            
            columns = columns["raw_data"]
        
        
        if not os.path.isfile(self.db_name):
            raise ValueError(f"database '{self.db_name}' does not exist")
        
        if "pc_time_end_measurement" not in columns:
            columns = ["pc_time_end_measurement"]+columns
        
        time_col = "pc_time_end_measurement"
        
        time_condition = " WHERE "
        if start_time != None:
            times_after = time_col+" > \""+str(start_time)+"\" and "
            time_condition += times_after
        
        if end_time != None:
            times_before =  time_col+" < \""+str(end_time)+"\" and "
            time_condition += times_before
        
        if time_condition == " WHERE ":
            time_condition = ""
        
        sql = "SELECT "+", ".join(columns)+" FROM raw_data" + time_condition
        
        load = SqliteDBLoad(self.db_name)
        result = load.load_sql(sql)
        
        result.columns = columns
        
        ser = Serialisation("numpy")
        result = ser.decode_dataframe(result, columns[1:])
        
        if len(parameter_columns) == 1:
            result = pd.DataFrame(np.stack(result[parameter_columns[0]]), 
                                  index=result["pc_time_end_measurement"], 
                                  columns=np.arange(300, 1101))
        
        return result
    
    
    def load_metadata(self, columns:[str]=[], 
                      condition:str="", 
                      raise_on_missing:bool=True
                      ) -> pd.DataFrame:
        """loads metadata from the database
        params:
            columns: list of column names to extract from database, if empty, returns all
            condition: sql that comes after the WHERE
                e.g: gobal_molar>50
        returns:
            dataframe with the requested values.
        """
        return self.load(columns, condition=condition, 
                         table="deployment_metadata", raise_on_missing=raise_on_missing)
    
    def load_accessory(self, columns:[str]=[], 
                       start_time:str=None, 
                       end_time:str=None,
                       condition:str="",
                       raise_on_missing:bool=True,
                       sort:bool=True,
                       timezone:str="+00:00"
                       ) -> pd.DataFrame:
        """runs load with table="accessory". more readable than using the parameter, 
        as accessory cant be mixed with non accessory
        """
        table_names = self.load_table_names()
        
        if "accessory_data" not in table_names.keys():
            raise ValueError("this database dosent have accessory data")
        
        if columns == []:
            columns = table_names["accessory_data"]
        
        return self.load(columns, "accessory_data", start_time, end_time, condition,
                         raise_on_missing, sort, True, timezone)
    
    
    def load_spectrum(self, column,
                      table:str=None, 
                      start_time:str=None, 
                      end_time:str=None,
                      condition:str="",
                      raise_on_missing:bool=True,
                      sort:bool=True,
                      timezone:str="+00:00"
                      ) -> pd.DataFrame:
        """loads a single spectrum from the database and returns it as a dataframe
        
        params:
            column: which column to load
            table: which table to load data from. Not necessary but can be useful to specify when multiple tables have the same column
            start_time, end_time: for only selecting data within a time period
                format: ISO-8601("2023-05-23 00:00:00")
                if you dont put in a full datetime, it will assume first timestamp,
                    e.g: 2023-05-23 will be read as 2023-05-23 00:00:00.
                    therefore, start_time is inclusive, end time is not
            condition: sql that comes after the WHERE in an sql statement.
                e.g: gobal_integral>50
                if you want even more control, use load_sql
            raise_on_missing: how to handle when columns arent found, raise or just load if possible
            sort: sorts the results from the database
            deserialise: wether or not to automatically deserialise spectral data
            timezone: the timezone that all the data will be converted to
        
        returns:
            one dataframe containing the spectral data in the requested column
                columns = wavelengths values; 300-1100
                index = time of measurement
        """
        if not isinstance(column, str):
            raise TypeError("column must be a string. if you want to load multiple columns, call this function multiple times")
        
        data = self.load(["pc_time_end_measurement", column], table, start_time, end_time, condition,
                         raise_on_missing, sort, timezone)
        if len(data)>0:
            if isinstance(data.loc[0, column], np.ndarray):
                print("spectrum")
                return pd.DataFrame(np.stack(data[column].values), index=data["pc_time_end_measurement"], columns=np.arange(300, 1101, 1))
            else:
                print("could not load spectrum, column was not a dataframe")
        else:
            print("could not load spectrum, no rows were returned from the dataframe")
        
        return pd.DataFrame()
    
        
    def load_sql(self, sql:str) -> pd.DataFrame:
        """loads data from a database using an sql query from the user
        
        params:
            sql: sql to select the desired data from the database
        
        returns:
            dataframe of the selected values
        """
        
        if not os.path.isfile(self.db_name):
            print("database: "+self.db_name+" does not exist")
            return None
        
        con = sqlite3.connect(self.db_name)
        cur = con.cursor()
        data = cur.execute(sql).fetchall()
        con.close()
        
        if data == []:
            return None
        return pd.DataFrame(data)
    
    
    def __guess_type(self, column_headers, columns):
        """when the user dosent explicitly pass which table the requested columns are in,
        this code guesses what type of table the requested columns are in, 
        as you cant return data from different length tables at the same time.
        the types are:
            "normal": spectral_data, system_data, precalculated_values.
                readings every minute, on the minute.
            "accessory": accessory_data.
                readings every 10 seconds
            "deployment": deployment_metadata.
                one reading every dataseries
        """
        _type = "normal"
        if "accessory_data" in column_headers.keys() or "deployment_metadata" in column_headers.keys():
            normal_column_headers = column_headers.copy()
            
            accessory_column_headers = deployment_column_headers = []
            if "accessory_data" in column_headers.keys():
                accessory_column_headers = column_headers["accessory_data"]
                del normal_column_headers["accessory_data"]
            
            if "deployment_metadata" in column_headers.keys():
                deployment_column_headers = column_headers["deployment_metadata"]
                del normal_column_headers["deployment_metadata"]
            
            
            
            normal_column_headers_list = []
            for key in normal_column_headers:
                normal_column_headers_list += normal_column_headers[key]

            ##### one liner to remove duplicates from a list while preserving the order
            normal_column_headers_list = list(dict.fromkeys(normal_column_headers_list).keys())
            

            ##### remove time from the no_accessory list because it is also in accessory_data
            #####   this was causing errors when time and an accessory_data value were both selected, as 
            #####   only the spectral data time was being returned
            ##### this could cause issues if a user wants only the spectral data time, as theres no way to specify
            ##### currently defaults to spectral data if only pc_time is requested
            if "pc_time_end_measurement" in normal_column_headers_list and len(columns) > 1:
                normal_column_headers_list.remove("pc_time_end_measurement")
            
            check_columns = columns.copy()
            if "pc_time_end_measurement" in check_columns:
                check_columns.remove("pc_time_end_measurement")
            
            
            has_normal_data = np.isin(check_columns, normal_column_headers_list).any()
            has_deployment_data = np.isin(check_columns, deployment_column_headers).any()
            has_accessory_data = np.isin(check_columns, accessory_column_headers).any()
             
            ##### there are duplicate column names in different tables
            ##### where there are columns in multiple tables being selected, 
            #####   check if all the columns can be found in one table, then pick that table
            ##### if not, raise
            if has_accessory_data and has_normal_data:
                ##### is there a column being requested that is only in one table
                only_accessory = False
                only_normal = False
                for col in check_columns:
                    if col in normal_column_headers_list and col not in accessory_column_headers:
                        only_normal = True
                    if col in accessory_column_headers and col not in normal_column_headers_list:
                        only_accessory = True
                ##### if all requested columns in accessory are also in a normal table
                #####   ignore accessory and vice versa
                if only_normal and not only_accessory:
                    has_accessory_data = False
                if only_accessory and not only_normal:
                    has_normal_data = False
                if not only_normal and not only_accessory:
                    has_accessory_data = False
            
            
            
            ##### do same check for deployment
            if has_deployment_data and has_normal_data:
                ##### is there a column being requested that is only in one table
                only_deployment = False
                only_normal = False
                for col in check_columns:
                    if col in normal_column_headers_list and col not in deployment_column_headers:
                        only_normal = True
                    if col in deployment_column_headers and col not in normal_column_headers_list:
                        only_deployment = True
                
                ##### if all requested columns in deployment are also in a normal table
                #####   ignore deployment and vice versa
                if only_normal and not only_deployment:
                    has_deployment_data = False
                if only_deployment and not only_normal:
                    has_normal_data = False
                if not only_normal and not only_deployment:
                    has_deployment_data = False
            
            num_tables = sum([has_normal_data, has_deployment_data, has_accessory_data])
            
            if num_tables > 1 and len(columns) > 1:
                raise Exception("You have requested data from incompatible tables (probably because they have unequal lengths). " + 
                                "Make multiple seperate requests")
            
            if num_tables == 1:
                if has_normal_data:
                    _type = "normal"
                if has_accessory_data:
                    _type = "accessory"
                if has_deployment_data:
                    _type = "deployment"
            
            
            if _type == "accessory":
                column_headers = {"accessory_data": column_headers["accessory_data"]}
            else:
                if "accessory_data" in column_headers.keys():
                    del column_headers["accessory_data"]
            
            if _type == "deployment":
                column_headers = {"deployment_metadata": column_headers["deployment_metadata"]}
            else:
                if "deployment_metadata" in column_headers.keys():
                    del column_headers["deployment_metadata"]
        
        return column_headers, _type


# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""

import sqlite3
import pandas as pd
import numpy as np
import time
import os
import uuid

import hsr1

class SqliteDBStore():
    def __init__(self, driver):
        self.db_name = driver.db_name
        self.driver = driver
    
    
    
    def store(self, dfs, precalculate=True):
        """stores data to the database
        
        This function should usually only be used straight after reading data from a dataset of .zip files, 
        and you can pass the output of that straight into the dfs parameter of this function
        
        params:
            spectral_data, system_data, deployment_metadata: dataframes containing the data that will be stored to the database
        """
        print("storing to "+self.db_name)
        
        driver = self.driver
        
        ##### unpacking dfs from list
        spectral_data, system_data, deployment_metadata, *accessory_data_list = dfs
        accessory_data = accessory_data_list[0] if len(accessory_data_list) >= 1 else pd.DataFrame()
        
        ##### changes the datatypes of all the uuids to strings, as sqlite dosent support uuids
        spectral_data, system_data, deployment_metadata = self.__sqlite_change_types([spectral_data, system_data, deployment_metadata])
        deployment_metadata = deployment_metadata.reset_index(drop=True)
        
        # TODO: more thorough check if db exists
        ##### checks if appending to a database or making new one, and if it already exists, checks to see if its the same deployment
        if driver.exists():
            db_load = hsr1.db.SqliteDBLoad(self.db_name)
            deployment_metadata = self.match_deployment_ids(deployment_metadata, db_load)
            deployment_metadata = self.match_dataseries_ids(deployment_metadata, db_load)
        else:
            deployment_metadata["deployment_id"] = str(uuid.uuid1())
            deployment_metadata["dataseries_id"] = str(uuid.uuid1())
        
        spectral_data, system_data, accessory_data = self.__add_dataseries_ids(deployment_metadata["dataseries_id"].iloc[0], 
                                                                               spectral_data, system_data, accessory_data)
        
        
        ##### only add new data values
        if driver.exists():
            db_load = hsr1.db.SqliteDBLoad(self.db_name)
            table_names = db_load.load_table_names()
            if "deployment_metadata" in table_names.keys() and "spectral_data" in table_names.keys() and "system_data" in table_names.keys():
                ##### generate a dictionary that matches dataseries id to deployment id
                ##### key: dataseries_id value: deployment_id
                dataseries_deployment_ids = db_load.load(["dataseries_id", "deployment_id"])
                deployment_ids = {}
                for i in range(len(dataseries_deployment_ids)):
                    deployment_ids[dataseries_deployment_ids.loc[i, "dataseries_id"]] = dataseries_deployment_ids.loc[i, "deployment_id"]
                for i in range(len(deployment_metadata)):
                    new_dataseries_id = deployment_metadata.loc[i, "dataseries_id"]
                    new_deployment_id = deployment_metadata.loc[i, "deployment_id"]
                    if new_dataseries_id not in deployment_ids.keys():
                        deployment_ids[new_dataseries_id] = new_deployment_id
                
                spectral_data = self.__check_for_duplicates(spectral_data, "spectral_data", deployment_ids, db_load)
                system_data = self.__check_for_duplicates(system_data, "system_data", deployment_ids, db_load)
                
                
                ##### check accessory_data for duplicates
                if len(accessory_data) >= 1:
                    accessory_data = self.__check_for_duplicates(accessory_data, "accessory_data", deployment_ids, db_load)
                
                
                if len(spectral_data) == 0:
                    print("all timestamps are duplicated, no data is being added")
        
        if not os.path.exists(self.db_name):
            os.makedirs(os.path.dirname(self.db_name), exist_ok=True)
        
        con = sqlite3.connect(self.db_name)
        
        normal_dfs = [spectral_data, system_data]
        normal_dfs_names = ["spectral_data", "system_data"]
        if len(accessory_data_list) >= 1:
            normal_dfs += [accessory_data]
            normal_dfs_names += ["accessory_data"]
            
        for i in range(len(normal_dfs)):
            if normal_dfs[i] is not None:
                self.store_dataframe(normal_dfs[i], normal_dfs_names[i])
        
        deployment_metadata.to_sql("deployment_metadata", con, if_exists="append", index=False)
        con.commit()
        con.close()
        
        if precalculate:
            mobile = False
            if deployment_metadata.loc[len(deployment_metadata)-1, "mobile"] == "1":
                mobile = True
            
            start_time = time.perf_counter()
            driver.add_precalculated_values(mobile=mobile, sample_ids_to_add=spectral_data["sample_id"].values, drop_existing=False)
            print("time to precalculate: ", time.perf_counter()-start_time)
    
    
    def store_raw(self, dfs, deployment_metadata):
        """stores raw data to the database
        
        params:
            dfs: tuple of dataframes, one dataframe per channel
            deployment_metadata: dataframe with one row, containing the deployment metadata for the dataseries being stored
        
        if no raw data already exists, creates a new table, "raw_data" which stores spectral data like in the spectral_data
        table, with each reading containing a blob that can be decoded to a spectral array
        """
        table_names = ["channel_"+str(i) for i in range(len(dfs))]
        
        big_df = pd.DataFrame()
        for i, df in enumerate(dfs):
            channel = pd.DataFrame(hsr1.db.Serialisation.listify_and_serialise_numpy(df), columns=[table_names[i]])
            channel = channel.reset_index()
            channel.columns = ["pc_time_end_measurement", table_names[i]]
            channel["pc_time_end_measurement"] = channel["pc_time_end_measurement"].astype(str)
            if len(big_df) == 0:
                big_df = channel
            else:
                big_df = big_df.merge(channel, on="pc_time_end_measurement")
        
        
        store = SqliteDBStore(self.driver)
        
        deployment_metadata = deployment_metadata.reset_index(drop=True)
        
        # TODO: more thorough check if db exists
        ##### checks if appending to a database or making new one, and if it already exists, checks to see if its the same deployment
        if self.driver.exists():
            db_load = self.driver.db_load
            store = SqliteDBStore(self.driver)
            deployment_metadata = store.match_deployment_ids(deployment_metadata, db_load)
            deployment_metadata = store.match_dataseries_ids(deployment_metadata, db_load)
        else:
            deployment_metadata["deployment_id"] = str(uuid.uuid1())
            deployment_metadata["dataseries_id"] = str(uuid.uuid1())
        
        store.store_dataframe(deployment_metadata, "deployment_metadata")
        store.store_dataframe(big_df, "raw_data") 
    
        
        

    def combine_database(self, new_db_name:str):
        """combines another database into the current database
        params:
            new_db_name: filepath to the database that will be merged into the current one
        
        """
        
        if not os.path.isfile(new_db_name):
            raise ValueError("databse to be merged: \""+new_db_name+" \" dosent exist")
        
        
        
        ##### avoid loading duplicates
        new_load = hsr1.db.SqliteDBLoad(new_db_name)
        
        new_metadata = new_load.load_metadata()
        metadata_to_store = new_metadata
        
        non_match_index_normal = None
        non_match_index_accessory = None
        
        existing_load = hsr1.db.SqliteDBLoad(self.db_name)
        
        new_tables = new_load.load_table_names()
        existing_tables = existing_load.load_table_names().keys()
        
        new_raw = False
        new_regular = False
        existing_raw = False
        existing_regular = False
        
        if "raw_data" in new_tables:
            new_raw = True
        if len(new_tables) >= 3:
            new_regular = True
        
        if "raw_data" in existing_tables:
            existing_raw = True
        if len(existing_tables) >= 3:
            existing_regular = True
        
        
        matched_metadata = self.match_deployment_ids(new_metadata, existing_load)
        metadata_to_store = matched_metadata
        
        
        if new_regular and existing_regular:
            ##### load all the dataseries_ids and deployment_ids from both tables and put into one dataframe
            existing_dataseries_deployment_ids = existing_load.load(["dataseries_id", "deployment_id"])
            new_dataseries_deployment_ids = matched_metadata[["dataseries_id", "deployment_id"]]
            dataseries_deployment_ids = pd.concat((existing_dataseries_deployment_ids, new_dataseries_deployment_ids))
            dataseries_deployment_ids = dataseries_deployment_ids.reset_index()
            
            deployment_ids = {}
            for i in range(len(dataseries_deployment_ids)):
                deployment_ids[dataseries_deployment_ids.loc[i, "dataseries_id"]] = dataseries_deployment_ids.loc[i, "deployment_id"]
            
            ##### dataframes with data from the database and data that is about to be added
            new_values = new_load.load(["pc_time_end_measurement", "dataseries_id"])
            
            existing_values = existing_load.load(["pc_time_end_measurement", "dataseries_id"])
            
            
            non_match_index_normal, num_dupes = self.find_matching_rows(existing_values, new_values, deployment_ids)
            print("skipping "+str(num_dupes)+" duplicated timestamps in spectral_data, system_data and precalculated_values")
            
            
            
            ##### load table names for existing and new table names
            existing_table_names = existing_load.load_table_names()
            new_table_names = new_load.load_table_names()
            ##### only need to check for dupes if both tables have accessory data
            if "accessory_data" in existing_table_names.keys() and "accessory_data" in new_table_names.keys():
                print("checking accessory_data for duplicate times")
                ##### dataframes with data from the database and data that is about to be added
                existing_values = existing_load.load_accessory(["pc_time_end_measurement", "dataseries_id"])
                new_values = new_load.load_accessory(["pc_time_end_measurement", "dataseries_id"])
                
                non_match_index_accessory, num_dupes = self.find_matching_rows(existing_values, new_values, deployment_ids)
                print("skipping "+str(num_dupes)+" duplicated timestamps in accessory_data")
            
            
            
            if len(non_match_index_normal) == 0 and non_match_index_accessory is not None and len(non_match_index_accessory) == 0:
                print("no data to merge: all times are duplicated")
            
            ##### sqlite ROWIDs are 1-based
            non_match_index_normal += 1
            non_match_index_normal = str(list(non_match_index_normal))
            non_match_index_normal = "(" + non_match_index_normal[1:]
            non_match_index_normal = non_match_index_normal[:-1] + ")"
            
            if non_match_index_accessory is not None:
                non_match_index_accessory += 1
                non_match_index_accessory = str(list(non_match_index_accessory))
                non_match_index_accessory = "(" + non_match_index_accessory[1:]
                non_match_index_accessory = non_match_index_accessory[:-1] + ")"
            
            
        
        
        con = sqlite3.connect(self.db_name)
        cur = con.cursor()
        
        cur.execute("ATTACH \"" +new_db_name+"\"AS new_db")
        cur.execute("PRAGMA database_list;")
        
        new_load = hsr1.db.SqliteDBLoad(new_db_name)
        table_names = new_load.load_table_names()
        
        for table_name in table_names.keys():
            ##### TODO: remove?  \/
            sql = "PRAGMA table_info("+table_name+");"
            cur.execute(sql)
            # column_headers = [column[1] for column in cur.fetchall()]
            column_headers = table_names[table_name]
            column_string = ", ".join(column_headers)
            string = "CREATE TABLE IF NOT EXISTS "+table_name+"("+column_string+")"
            cur.execute(string)
            if table_name != "deployment_metadata":
                if non_match_index_normal is not None and table_name in ["spectral_data", "system_data", "precalculated_values"]:
                    cur.execute("INSERT INTO main."+table_name+" SELECT * FROM new_db."+table_name+" WHERE ROWID in "+non_match_index_normal)
                elif non_match_index_accessory is not None and table_name == "accessory_data":
                    cur.execute("INSERT INTO main."+table_name+" SELECT * FROM new_db."+table_name+" WHERE ROWID in "+non_match_index_accessory)
                else:
                    cur.execute("INSERT INTO main."+table_name+" SELECT * FROM new_db."+table_name)
        
        
        
        con.commit()
        con.close()
        self.store_dataframe(metadata_to_store, "deployment_metadata")

    
    def combine_database_folder(self, folder, delete=False):
        files = os.listdir(folder)
        databases = []
        for file in files:
            if file[-2:] == "db":
                databases.append(file)
        
        for db_name in databases:
            self.combine_database(folder+db_name)
            if delete:
                os.remove(folder+db_name)
            
    
    
    
    def drop_table(self, table_name:str):
        """drops a given table from the database
        
        params:
            table_name: the name of the table to be dropped
        """
        if not os.path.isfile(self.db_name):
            print("database does not exist")
            return None
        
        con = sqlite3.connect(self.db_name)
        cur = con.cursor()
        cur.execute("DROP TABLE IF EXISTS "+table_name)
        con.commit()
        con.close()
    
    
    
    
    def match_deployment_ids(self, deployment_metadata:pd.DataFrame, db_load):
        """loads the deployment metadata from the database and checks if the deployment metadata being stored and matches the deployment ids if it is
        params:
            deployment_metadata: the new dataframe being added to the database
            db_load: a db_load object that can load from the database
        returns:
            deployment_metadata: the original dataframe with updated deployment_ids if required
        
        iterates through each row of deployment data that will be added, and checks it against each row of deployment_metadata already in the database,
        if the deployment data(dosent check dataseries data) matches, copies the preexisting deployment_id to the new data
        this allows grouping all data from a specific deployment, when the deployment id isnt included in the deployment's config file'
        """
        # TODO: get this from deployment file
        # subset of columns that refer to the deployment, rather than the dataseries
        deployment_columns = ["data_structure_id", 
                              "deployment_description", 
                              "sensor_type", 
                              "sensor_id", 
                              "camera_id", 
                              "owner_contact", 
                              "operator_contact", 
                              "license", 
                              "license_reference",
                              "deployment_id"]
        
        deployment_metadata = self.__match_deployment_metadata_ids(deployment_metadata, deployment_columns, "deployment_id", db_load)
        
        return deployment_metadata

    
    def match_dataseries_ids(self, deployment_metadata:pd.DataFrame, db_load):
        dataseries_columns = ["mobile",
                              "processing_level",
                              "timezone",
                              "location_name",
                              "default_latitude",
                              "default_longitude",
                              "default_elevation",
                              "dataseries_id",
                              "platform_id",
                              "software_id",
                              "calibration_time",
                              "calibration_comment",
                              "camera_calibration_file",
                              "spectrometer_calibration_file",
                              "start_time",
                              "integration_time",
                              "gain",
                              "hdr",
                              "spectrometer_sampling_period",
                              "spectrometer_average_period",
                              "spectrometer_burst_number",
                              "wavelengths",
                              "aux_sampling_period",
                              "aux_average_period"]
        
        deployment_metadata = self.__match_deployment_metadata_ids(deployment_metadata, dataseries_columns, "dataseries_id", db_load)
        
        return deployment_metadata
    
    def store_dataframe(self, df:pd.DataFrame, table_name:str):
        """stores one dataframe to a table in the database
        
        params:
            df: the dataframe being stored
            table_name: the name of the table it will be stored to
        """
        con = sqlite3.connect(self.db_name)
        cur = con.cursor()
        
        question_marks = ["?" for i in range(len(df.columns))]
        question_marks_string = ", ".join(question_marks)
        
        column_string = ", ".join(df.columns.to_list())
        
        create_table_sql = "CREATE TABLE IF NOT EXISTS "+table_name+"("+column_string+")"
        cur.execute(create_table_sql)
        
        sql = "INSERT INTO "+table_name+" VALUES("+question_marks_string+")"
        cur.executemany(sql, df.values)
        
        con.commit()
        con.close()
    
    
    def __match_deployment_metadata_ids(self, df, columns, id_type:str, db_load):
        try:
            existing_deployment_metadata = db_load.load_metadata(columns)
        except KeyError:
            print("no deployment_metadata found in existing table")
            return df
        new_df = df[columns]
        
        # loops through each row of new deployment data (can be multiple, e.g: copying across databases)
        #   and through checks it against each row of existing deployment data from the database
        #   dosent check deployment_id column because that will be different
        for i, deployment in enumerate(new_df.drop(id_type, axis=1).iterrows()):
            found = False
            for j, existing_deployment in enumerate(existing_deployment_metadata.drop(id_type, axis=1).iterrows()):
                if deployment[1].equals(existing_deployment[1]):
                    found = True
                    print("deployment metadata matches, using same deployment id")
                    df.loc[i,id_type] = existing_deployment_metadata.loc[j,id_type]
                    break
            if not found:
                print("no deployment id found, generating new one")
                df.loc[i,id_type] = str(uuid.uuid1())
        
        return df
    
    def __add_dataseries_ids(self, dataseries_id, spectral_data, system_data, accessory_data=None):
        """adds a given dataseries id to given dataframes"""
        if spectral_data is not None:
            spectral_data["dataseries_id"] = dataseries_id
        if system_data is not None:
            system_data["dataseries_id"] = dataseries_id
        if accessory_data is not None and len(accessory_data) > 0:
            accessory_data["dataseries_id"] = dataseries_id
            
        return spectral_data, system_data, accessory_data
    
    def __sqlite_change_types(self, dataframes:[pd.DataFrame]) -> [pd.DataFrame()]:
        """change the type of all the uuids to strings
        params:
            dataframes: list of dataframes to be converted
                [spectral_data, system_data, deployment_metadata]
        returns:
            dataframes: list of converted dataframes, same format as parameter
        """
        
        uuid_columns = ["sample_id", "dataseries_id", "deployment_id"]
        
        output_dataframes = []
        for df in dataframes:
            if df is None:
                output_dataframes.append(None)
            else:
                for column in uuid_columns:
                    if column in df.columns:
                        df[column] = df[column].astype(str)
                output_dataframes.append(df)
        return output_dataframes
    
    
    def __check_for_duplicates(self, data, table_name, deployment_ids, db_load):
        """
        params:
            deployment_ids: a dictionary that can convert dataseries_ids
              to deployment_ids. dataseries_ids are the keys and deployment_ids are the values.
        """
        print("checking for duplicates")
        existing_values = None
        ##### dataframes with data from the database and data that is about to be added
        try:
            existing_values = db_load.load(["pc_time_end_measurement", "dataseries_id"], table=table_name)
        except ValueError:
            ##### if there is no data in the database for this table, no need to check for dupes, just return original data
            return data
        
        new_values = data[["pc_time_end_measurement", "dataseries_id"]].copy()
        new_values["pc_time_end_measurement"] = pd.to_datetime(new_values["pc_time_end_measurement"]).dt.tz_convert("+00:00")
        
        existing_values["pc_time_end_measurement"] = pd.to_datetime(existing_values["pc_time_end_measurement"]).dt.tz_convert(None)
        new_values["pc_time_end_measurement"] = pd.to_datetime(new_values["pc_time_end_measurement"]).dt.tz_convert(None)
        
        
        ##### find index of all non-duplicated values
        non_match_index, num_dupes = self.find_matching_rows(existing_values, new_values, deployment_ids)
        print("skipping "+str(num_dupes)+" duplicated timestamps in "+table_name)        
        
        ##### filter for all valid readings
        data = data.iloc[non_match_index]
        
        return data
    
    def find_matching_rows(self, existing, new, deployment_ids):
        ##### replace dataseries_id with deployment id and rename
        existing = existing.replace(deployment_ids).rename(columns={"dataseries_id":"deployment_id"})
        new = new.replace(deployment_ids).rename(columns={"dataseries_id":"deployment_id"})
        
        ##### selecting for same deployment
        deployment_id = new.loc[1, "deployment_id"]
        id_existing = existing.loc[existing["deployment_id"] == deployment_id, "pc_time_end_measurement"]
        id_new = new.loc[new["deployment_id"] == deployment_id, "pc_time_end_measurement"]
        
        if len(id_existing) == 0:
            print("new data has different deployment_metadata, creating a new d")
        
        id_existing = pd.to_datetime(id_existing)
        id_new = pd.to_datetime(id_new)
        
        
        existing_limits = (id_existing.loc[0],
                           id_existing.loc[len(id_existing)-1])
        new_limits = (id_new.loc[0],
                      id_new.loc[len(id_new)-1])
        
        existing_intersection = np.logical_and(id_existing >= new_limits[0],
                                               id_existing <= new_limits[1])
        new_intersection = np.logical_and(id_new >= existing_limits[0],
                                          id_new <= existing_limits[1])
        
        trimmed_existing_values = id_existing.loc[existing_intersection].reset_index(drop=True)
        trimmed_new_values = id_new.loc[new_intersection].reset_index(drop=True)
        
        matches = np.isin(trimmed_new_values, trimmed_existing_values)
        
        invalid = new_intersection
        invalid[new_intersection] = matches
        
        
        non_match_index = id_new.index[np.logical_not(invalid)]
        
        num_dupes = sum(matches)
        
        return non_match_index, num_dupes
    
    


# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""

import pickle
import numpy as np
import array
import pandas as pd

class Serialisation:
    
    def __init__(self, method):
        self.method = method
    
    def __encode_array(self, data):
        """encodes a list of values using pythons array.array module"""
        # parameter can be 2d array or pd.Series of lists
        return [array.array("f", x).tobytes() for x in data]

    def __decode_array(self, binary):
        """decodes a byte array using pythons array.array module"""
        # parameter can be 2d array or pd.Series of lists
        decoded_data = []
        for data_array in binary:
            new_array = array.array("f") 
            new_array.frombytes(data_array)
            decoded_data.append(new_array.tolist())
        return decoded_data


    def __encode_pickle(self, data):
        """encodes a list of values using pythons pickle module"""
        # parameter can be 2d array or pd.Series of lists
        return [pickle.dumps(x) for x in data]

    def __decode_pickle(self, binary):
        """decodes a pickled object into a list of values"""
        # parameter can be 2d array or pd.Series of lists
        return [pickle.loads(x) for x in binary]


    def __encode_numpy(self, data):
        """encodes a list of values to bytes using numpy's to_bytes method"""
        # parameter can be 2d array or pd.Series of lists
        return [np.array(x).tobytes() for x in data]

    def __decode_numpy(self, binary):
        """decodes a bytearray to a numpyarray of values"""
        decoded_data = []
        for data_array in binary:
            decoded_data.append(np.frombuffer(data_array))
        return decoded_data
    
    
    def listify_and_serialise_numpy(df):
        """converts a dataframe with many columns into a series with a byte representation of an array of each row"""
        return pd.Series([np.array(x[1]).tobytes() for x in df.iterrows()], df.index)
        


    def encode_dataframe(self, spectral_data_df:pd.DataFrame):
        """encodes the spectral data of a dataframe using the instance variable method"""
        if self.method == "numpy":
            spectral_data_df["global_spectrum"] = self.__encode_numpy(spectral_data_df["global_spectrum"])
            spectral_data_df["diffuse_spectrum"] = self.__encode_numpy(spectral_data_df["diffuse_spectrum"]) 
        elif self.method == "pickle":
            spectral_data_df["global_spectrum"] = self.__encode_pickle(spectral_data_df["global_spectrum"])
            spectral_data_df["diffuse_spectrum"] = self.__encode_pickle(spectral_data_df["diffuse_spectrum"])        
        else:
            spectral_data_df["global_spectrum"] = self.__encode_array(spectral_data_df["global_spectrum"])
            spectral_data_df["diffuse_spectrum"] = self.__encode_array(spectral_data_df["diffuse_spectrum"]) 
        return spectral_data_df

    def decode_dataframe(self, spectral_data_df:pd.DataFrame, columns:[str]=["global_spectrum", "diffuse_spectrum"]):
        """decodes a subset of the spectral data in a dataframe
        params:
            spectral_data_df: the dataframe containing serialised spectral data
        columns:
            the specific columns in the dataframe that contain spectral data and are to be deserialised
            (saves time if e.g. only global spectrum is required)
        """
        if type(columns) == str:
            columns = [columns]
        
        if self.method == "numpy":
            for column in columns:
                spectral_data_df[column] = self.__decode_numpy(spectral_data_df[column])
        elif self.method == "pickle":
            for column in columns:
                spectral_data_df[column] = self.__decode_pickle(spectral_data_df[column])      
        else:
            for column in columns:
                spectral_data_df[column] = self.__decode_array(spectral_data_df[column])
        return spectral_data_df



# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""

import numpy as np
import matplotlib.pyplot as plt

from hsr1.utils.spectrum.pixelSpectrum import PixelSpectrum


class WeightedSpectrumSegment:
    def __init__(self, ref_spectrum, measured_spectrum, centre, distance, old_centre_point):
        self.ref_spectrum = ref_spectrum
        self.measured_spectrum = measured_spectrum
        
        self.distance = distance
        
        self.old_centre_point = old_centre_point
        
        self.left = centre-distance
        self.right = centre+distance
        
        left_padding_dist = max(0, self.left)
        right_padding_dist = max(0, len(ref_spectrum)-(self.right))
        
        error = 0
        if self.left <0:
            error = abs(self.left)
            self.left = 0
        
        self.segment = ref_spectrum[self.left:self.right]
        
        if error > 0:
            self.segment = np.concatenate((self.segment, np.zeros(error)))
            right_padding_dist -= error
        
        self.windowed_spectrum = np.concatenate((np.zeros(left_padding_dist), self.segment, np.zeros(right_padding_dist)))
        if max(self.windowed_spectrum) > 0:
            self.windowed_spectrum = self.windowed_spectrum/max(self.windowed_spectrum)
        
        self.centre = centre
        
    
    
    def plot(self, c0, c1, fit_quality, calibrated_spectrum):
        plt.figure(figsize=(20,10))
        plt.suptitle(str(c0) + "\n" + str(c1) + f"\n{fit_quality:.2e}")
        plt.plot(self.windowed_spectrum)
        plt.plot(calibrated_spectrum, label="calibrated")
        plt.axvline(self.centre, c="black")
        plt.axvline(self.centre+self.distance, c="gray")
        plt.axvline(self.centre-self.distance, c="gray")
        plt.legend()
        plt.show()
        
    
    def find_best_calibration(self, c0_init, c1_init, 
                              c0_lim, c1_lim,
                              iterations=20,
                              plot=False, debug=True):
        """finds the best calibration for this segment
        params:
            c0_init: the starting c0 value
            c1_init: the starting c1 value
            c0_lim: how far to the left and right to shift
            c1_lim: how much to stretch and squash by
            iterations: How many different stretch values to try within the limit (every integer shift is tried within the limits)
            plot: plot every single set of values and the calibrated spectrum they produce
            debug: return debug info
        returns:
            poly: polynomial calibration for this segment
            all_attempts: a 2d array of all the fit quality for all the tested values
        """
        ps = PixelSpectrum()
        
        ##### when a neighbouring segment squashes the spectrum, assuming it
        #####   has an accurate calibration for that segment, will be a good starting point for this segment.
        #####   not only can we start by using the c0 and c1 values from the previous segment,
        #####   but we can also apply a slight extra shift to account for this segment's
        #####   centre point being slightly off from where the previous segment's calibration would suggest it might be.
        #####   by applying a shift of previous_stretch-1 * change_in_centre_points,
        #####   we can get a more accurate guess of the starting shift, and can therefore have a more restricted c0_lim
        
        adjustment = (c1_init-1)*(self.centre-self.old_centre_point)
        c0_init += adjustment
        
        results = {}
        all_attempts = []
        ##### c1 is applied before c0, because a stretch takes longer than a shift so this reduces the total time
        for c1 in np.linspace(c1_init-c1_lim, c1_init+c1_lim, iterations):
            ##### apply the new c1
            stretched_spectrum = ps.apply_origin_wavelength_calibration(self.measured_spectrum, (0, c1, 0, 0), self.centre)
            shift_results = []
            for c0 in np.arange(c0_init-c0_lim, c0_init+c0_lim):
                ##### apply the shift
                calibrated_spectrum = ps.apply_constant_wavelength_calibration(stretched_spectrum, c0)
                
                if len(calibrated_spectrum[self.left:self.right]) > 0 and max(calibrated_spectrum[self.left:self.right]) > 0:
                    ##### normalise
                    calibrated_spectrum = calibrated_spectrum/max(calibrated_spectrum[self.left:self.right])
                
                
                ##### calculates fit_quality
                fit_quality = np.sum(calibrated_spectrum*self.windowed_spectrum)
                results[(c0, c1)] = fit_quality
                shift_results.append(fit_quality)
                
                if plot:
                    self.plot(c0, c1, fit_quality, calibrated_spectrum)
            
            all_attempts.append(np.array(shift_results))
        all_attempts = np.array(all_attempts)
        
        ##### pick the polynomial that has the best fit quality
        best_fit = max(results, key=results.get)
        
        if debug:
            return [best_fit[0], best_fit[1], 0, 0], all_attempts
        
        return [best_fit[0], best_fit[1], 0, 0]
                
    


# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""
import time

import uuid

import pandas as pd
import numpy as np


from hsr1.utils.config import Config

class ReformatData():
    def reformat_data(self, dataframes:[pd.DataFrame], deployment_metadata_file_path:str, gps_type=None):
        """combines various methods to reformat all the data necessary for storing to the database
        params:
            dataframes: list of 6 dataframes: [ed, eds, summary, hdr, ind_ch, gps]
            deployment_metadata_file_path: path to the deployment metadata .ini file
            gps_type: the type of gps file, either "gps", or accessory"
        returns:
            spectral_data, precalculated_values, system_data, deployment_metadata
        """
        print("reformatting data")
        
        if gps_type is None:
            columns = dataframes[4].columns
            if len(columns) >= 12:
                print("no gps_type passed, assuming accessory")
                gps_type = "accessory"
            else:
                print("no gps_type passed, assuming gps")
                gps_type = "gps"
        
        deployment_metadata = self.reformat_deployment_metadata(deployment_metadata_file_path)
        spectral_data = self.reformat_spectral_data(dataframes[:3])
        system_data = None
        accessory_data = None
        if dataframes[4] is not None:
            accessory_data = None
            if gps_type == "accessory":
                accessory_data = self.reformat_accessory_data(dataframes[4])
            
            system_data = self.reformat_system_data(dataframes[4], gps_type, accessory_data, spectral_data.index)
            
            
        
        if system_data is not None:
            system_data = self.match_sample_ids(spectral_data, system_data, "sample_id")
            system_data = self.reset_index(system_data, "pc_time_end_measurement")
        
        spectral_data = self.reset_index(spectral_data, "pc_time_end_measurement")
        
        tz_string = deployment_metadata.iloc[0]["timezone"]
        timezone, timedelta = self.calculate_timezone(tz_string)
        
        
        default_location = [deployment_metadata["default_longitude"].iloc[0],  
                            deployment_metadata["default_latitude"].iloc[0],
                            deployment_metadata["default_elevation"].iloc[0]]
        for i in range(len(default_location)):
            if default_location[i] == "''":
                default_location[i] = np.nan
            default_location[i] = float(default_location[i])
        
        if system_data is not None:
            invalid_gps = system_data["gps_status"].str.slice(stop=1) == "V"
            system_data.loc[invalid_gps, "gps_longitude"] = default_location[0]
            system_data.loc[invalid_gps, "gps_latitude"] = default_location[1]
            system_data.loc[invalid_gps, "gps_altitude"] = default_location[2]
        
        dfs = [spectral_data, system_data, accessory_data] if gps_type == "accessory" else [spectral_data, system_data]
        
        for df in dfs:
            if df is not None:
                # df["pc_time_end_measurement"] = (pd.to_datetime(df["pc_time_end_measurement"])+timedelta).astype(str)
                df["pc_time_end_measurement"] = df["pc_time_end_measurement"].astype(str)+timezone
        
        dfs = [spectral_data, system_data, deployment_metadata, accessory_data] if gps_type == "accessory" else [spectral_data, system_data, deployment_metadata]
        return tuple(dfs)
    

    def reformat_deployment_metadata(self, file_path:str) -> pd.DataFrame:
        """reformats the deployment metadata found in an .ini file
        params:
            file_path: the location of the deployment metadata .ini file
        returns:
            df: the database version of the deployment metadata file
        
        storing this as a one line dataframe is very inneficient, but the sizes are so small,
            it doesent have an impact on performance, so maybe its best to keep it consistent with the other tables
        """
        print("reformatting deployment metadata")
        config_reader = Config(file_path)
        params = config_reader.read_section("deployment") | config_reader.read_section("dataseries")
        
        if not "mobile" in params.keys():
            params["mobile"] = False
        if not "location_name" in params.keys():
            params["location_name"] = ""
        
        self.dataseries_id = params["dataseries_id"]
        df = pd.DataFrame.from_records(params, index = ["1"])
        
        return df
    
    
    def reformat_spectral_data(self, dataframes:[pd.DataFrame]) -> pd.DataFrame:
        """reformats spectral data from the format in the txt files to the database format
        params:
            dataframes: list of dataframes, [ed, eds, summary]
        returns:
            spectral_data: dataframe of same structure as database
        """
        print("reformatting spectral data")
        ed:pd.DataFrame = dataframes[0]
        eds:pd.DataFrame = dataframes[1]
        summary:pd.DataFrame = dataframes[2]
        
        # round times to the nearest minute
        ed = self.__round_times(ed, "min")
        eds = self.__round_times(eds, "min")
        summary = self.__round_times(summary, "min")
        
        # select all the rows in the dataframe that are not duplicates
        ed = ed[~ed.index.duplicated()]
        eds = eds[~eds.index.duplicated()]
        summary = summary[~summary.index.duplicated()]
        
        ed = self.listify(ed)
        eds = self.listify(eds)
        
        
        spectral_data = pd.DataFrame()
        
        spectral_data["global_spectrum"] = ed
        spectral_data["diffuse_spectrum"] = eds
        spectral_data = spectral_data.merge(summary, left_index=True, right_index=True)
        
        spectral_data["dataseries_id"] = ""
        
        # spectral_data = spectral_data.reset_index()
        spectral_data = spectral_data.rename(columns={#"index":"pc_time_end_measurement",
                                                      "Total W": "global_integral",
                                                      "Diffuse W": "diffuse_integral",
                                                      "Total Molar": "global_molar",
                                                      "Diffuse Molar": "diffuse_molar",
                                                      "Temp": "camera_temp"})
        
        # generate a unique sample ID for each row that will be matched to the system data table
        spectral_data["sample_id"] = [uuid.uuid1() for i in range(len(spectral_data))]
        
        # reposition the data to be in the correct order
        spectral_data = spectral_data[["sample_id",
                                       "dataseries_id", 
                                       "global_spectrum", 
                                       "diffuse_spectrum", 
                                       "global_integral", 
                                       "diffuse_integral", 
                                       "global_molar", 
                                       "diffuse_molar", 
                                       "camera_temp"]]
        return spectral_data
        
        
    
    def reformat_system_data(self, gps:pd.DataFrame, gps_type:str, accessory_data:pd.DataFrame=None, spectral_timestamps:pd.Series=None) -> pd.DataFrame:
        """reformats data from the gps text file to the database format
        params:
            gps: dataframe containing gps data
            gps_type: type of gps file used, "gps", or "accessory"
        returns:
            gps: reformatted dataframe in the correct format for the database
        """
        if gps_type == "gps":
            gps = self.reformat_system_data_gps(gps)
        elif gps_type == "accessory":
            # gps = self.reformat_system_data_accessory(gps)
            if accessory_data is None:
                raise ValueError("if gps_type is accessory, accessory_data must be passed")
            gps = self.reformat_system_data_reformatted_accessory(accessory_data, spectral_timestamps)
        else:
            print("gps_type not recognised")
        
        ##### reorder gps columns so it is consistent
        if "pc_time_end_measurement" in gps.columns:
            gps = gps.set_index("pc_time_end_measurement")
        gps = gps[["gps_time",
                   "gps_status",
                   "gps_latitude",
                   "gps_longitude",
                   "gps_altitude",
                   "pressure",
                   "baro_temp",
                   "rh",
                   "rh_temp",
                   "dataseries_id"]]
        return gps
    
    def reformat_system_data_gps(self, gps:pd.DataFrame) -> pd.DataFrame:
        """reformats data from the gps text file to the database format
        params:
            gps: dataframe containing gps data
        returns:
            gps: reformatted dataframe in the correct format for the database
        """
        print("reformatting gps data")
        
        if len(gps.columns) > 12:
            raise ValueError("trying to reformat gps data, too many columns in dataframe")
        
        gps = self.__round_times(gps, "min")
        
        gps = gps.rename(columns={"pc_time": "pc_time_end_measurement",
                                  "GPS Time": "gps_time",
                                  "Status": "gps_status",
                                  "Latitude": "gps_latitude",
                                  "Longitude": "gps_longitude",
                                  "Altitude": "gps_altitude",
                                  "Pressure": "pressure",
                                  "BaroTemp": "baro_temp",
                                  "RH": "rh",
                                  "RHTemp": "rh_temp"})
        
        gps = gps.drop(["HDOP", "TP0", "TP1"], axis = 1)
        
        gps["dataseries_id"] = ""
        
        return gps
    
    # def reformat_system_data_accessory(self, gps):
    #     """reformats data from the accessory text file to the database format
    #     params:
    #         gps: dataframe containing gps data
    #     returns:
    #         gps: reformatted dataframe in the correct format for the database
    #     """
    #     print("reformatting accessory data")
        
    #     if len(gps.columns) <= 12:
    #         raise ValueError("trying to reformat accessory data, too few columns in dataframe")
        
    #     gps = gps.drop(gps[gps["0 TimeYYYY"] == 0].index)
        
    #     gps = self.__round_times(gps, "min")
        
    #     ##### merge the gps time columns into one time column
    #     gps_date_cols = ["9 GPSYYYY", "10 GPSMonth", "11 GPSDay"]
    #     gps_time_cols = ["12 GPSHour", "13 GPSMin", "14 GPSSec"]
    #     gps[gps_date_cols] =  gps[gps_date_cols].astype(int).astype(str)
    #     gps[gps_time_cols] =  gps[gps_time_cols].astype(int).astype(str)
        
    #     ##### this line is the (speed) problem \/
    #     gps["gps_time"] = gps[gps_date_cols].agg("-".join, axis=1) + " " + gps[gps_time_cols].agg(":".join, axis=1)

    #     ##### where gps cuts out, use pc_time to stop errors
    #     gps["gps_time"] = gps["gps_time"].where(gps["gps_time"].str.slice(0,0) == "0", gps.index)
        
    #     gps["gps_time"] = pd.to_datetime(gps["gps_time"])
        
    #     gps = self.reset_index(gps, "pc_time_end_measurement")
        
    #     gps["gps_status"] = ""
    #     gps.loc[gps["18 GPSAge"] < 10, "gps_status"] = "A"
        
    #     new_gps = gps[["pc_time_end_measurement", "gps_time", "gps_status", "15 Latitude", "16 Longitude", "17 Altitude", "36 Pressure", "34 T-Baro", "35 RH%", "33 T-RH"]].copy()
        
        
    #     new_gps = new_gps.rename(columns={"15 Latitude":"gps_latitude",
    #                               "16 Longitude":"gps_longitude",
    #                               "17 Altitude":"gps_altitude",
    #                               "36 Pressure":"pressure",
    #                               "34 T-Baro":"baro_temp",
    #                               "35 RH%":"rh",
    #                               "33 T-RH":"rh_temp"})
        
    #     new_gps = new_gps.reset_index(drop=True)
        
        
    #     ##### averages rows to the nearest minute
    #     new_gps["pc_time_end_measurement"] = new_gps["pc_time_end_measurement"].round("min")
    #     ##### to get time from the end of the period:
    #     merged_gps = new_gps.groupby(new_gps["pc_time_end_measurement"])[["pc_time_end_measurement", "gps_time", "gps_latitude", "gps_longitude", "gps_altitude", "pressure", "baro_temp", "rh", "rh_temp"]].mean()
    #     # TODO: how to merge gps status(first? keep worst?)
    #     merged_gps["gps_status"] = new_gps[["pc_time_end_measurement", "gps_status"]].groupby(new_gps["pc_time_end_measurement"]).first()["gps_status"]
        
    #     merged_gps["dataseries_id"] = ""
        
    #     merged_gps["gps_time"] = merged_gps["gps_time"].astype(str)
        
    #     merged_gps = merged_gps.drop("pc_time_end_measurement", axis=1)
        
        
        
    #     return merged_gps
    
    
    def reformat_system_data_reformatted_accessory(self, accessory:pd.DataFrame, spectral_timestamps:pd.Series=None, spectral_sample_ids:pd.Series=None, aux_average_period=60):
        """averages accessory data to the same time period as spectral data,
        this will be stored in the system_data table
        
        params:
            accesssory: the raw accessory data.
            spectral_timestamps: all the timestamps where there are spectral readings.
        returns:
            averaged_system_data: ddataframe of the reformatted data.
        """
        print("reformatting accessory data")
        ##### filter out all the readings that arent within the aux_average_period, as those shouldnt be averaged
        
        
        system_data = accessory[["pc_time_end_measurement",
                               "gps_time",
                               "Latitude",
                               "Longitude",
                               "Altitude",
                               "Pressure",
                               "T_Baro",
                               "RH",
                               "T_RH"]]
        system_data = system_data.rename(columns = {"Latitude":"gps_latitude",
                                                    "Longitude":"gps_longitude",
                                                    "Altitude":"gps_altitude",
                                                    "Pressure":"pressure",
                                                    "T_Baro":"baro_temp",
                                                    "RH":"rh",
                                                    "T_RH":"rh_temp"})
        
        system_data["gps_status"] = "V"
        system_data.loc[accessory["GPSAge"] < 10, "gps_status"] = "A"
        
        
        if spectral_timestamps is None:
            print("no spectral_timestamps provided to average to, not averaging system_data")
            return system_data
        
        system_timestamps = pd.to_datetime(system_data["pc_time_end_measurement"]).astype("int64").values
        spectral_timestamps_int = pd.to_datetime(spectral_timestamps).astype("int64").values
        
        
        # #### split into chunks, one per day? limits the n^2edness
        # #### have to make sure no data is lost at the splits
        # start_time = time.perf_counter()
        # time_until_next_spectral = (spectral_timestamps_int[:, np.newaxis] - system_timestamps).astype(float)
        # time_until_next_spectral = time_until_next_spectral/10**9
        # time_until_next_spectral[time_until_next_spectral < 0] = np.inf
        # time_until_closest_spectral = np.min(time_until_next_spectral, axis=0)
        # print("big array time: " +str(time.perf_counter()-start_time))
        
        
        
        # start_time = time.perf_counter()
        # spectral_timestamps_int_editable = spectral_timestamps_int
        # time_until_closest_spectral = np.full(len(system_timestamps), np.inf)
        # ##### loops through every value which system_timestamps is less than the max spectral timestamp
        # #####   to ignore values that will never be averaged into spectral_timestamp and will cause errors
        # for i in range(np.sum(system_timestamps<=np.max(spectral_timestamps_int))):
        #     system_timestamp = system_timestamps[i]
            
        #     # spectral_timestamps_int_editable = spectral_timestamps_int_editable[spectral_timestamps_int_editable>=system_timestamp]
        #     # time_until_next_spectral = spectral_timestamps_int_editable[0]-system_timestamp
            
        #     ##### arg of first value that is greater than current. this will be the closest
        #     greater_than_current_pointer = np.argmax(spectral_timestamps_int_editable>=system_timestamp)
            
        #     #####convert to elementwise subraction after?
        #     time_until_next_spectral = spectral_timestamps_int_editable[greater_than_current_pointer]-system_timestamp
        #     time_until_closest_spectral[i] = time_until_next_spectral
        
        # time_until_closest_spectral = np.array(time_until_closest_spectral)/10**9
        # print("for loop time: " +str(time.perf_counter()-start_time))
        
        
        # within_average_period = time_until_closest_spectral < float(aux_average_period)
        

        # system_data = system_data[within_average_period].reset_index(drop=True)
        # system_timestamps = system_timestamps[within_average_period]
        
        str_cols = ["pc_time_end_measurement", "gps_time", "gps_status"]
        float_cols = ~np.isin(system_data.columns, str_cols)
        floats = system_data.loc[:, float_cols].astype(float)
        system_data.loc[:, float_cols] = floats
        
        
        agg_dict = {col:"last" for col in system_data.columns}
        
        for col in ["gps_latitude", "gps_longitude", "gps_altitude", "pressure", "baro_temp", "rh", "rh_temp"]:
            agg_dict[col] = "mean"
        averaged_system_data = system_data.groupby(spectral_timestamps_int[np.clip(np.digitize(system_timestamps, spectral_timestamps_int, right=True), 0, len(spectral_timestamps)-1)])#.agg(agg_dict)
        averaged_system_data = averaged_system_data.agg(agg_dict)
        
        int_to_timestamp = pd.DataFrame(index=spectral_timestamps_int)
        int_to_timestamp["pc_time_end_measurement"] = spectral_timestamps.values
        averaged_system_data["pc_time_end_measurement"] = int_to_timestamp.loc[averaged_system_data.index]
        
        ##### filter out readings that are more than aux_average_period from the previous reading.
        #####   filters the reading after big jumps
        time_since_last_reading = np.diff(pd.to_datetime(averaged_system_data["pc_time_end_measurement"]).astype("int64").values/10**9)
        normal_gaps = time_since_last_reading <= aux_average_period
        normal_gaps = np.array([1] + list(normal_gaps)).astype(bool)
        averaged_system_data = averaged_system_data.loc[normal_gaps, :]
        
        averaged_system_data = averaged_system_data.reset_index(drop=True)
        
        averaged_system_data[str_cols] = averaged_system_data[str_cols].astype(str)
        
        
        averaged_system_data["dataseries_id"] = ""
        
        return averaged_system_data
    
    
    def reformat_accessory_data(self, accessory):
        """changes column headings on accessory data to prepare it for storing to the database
        params:
            accessory: df containing the accessory data
        returns:
            accessory: reformatted df containing accessory data
        
        note: there is a wierd bug with the tara dataset where halfway through, the name of column 51
              from "51 Control flags" to "51 Control flags2"
        """
        
        ##### merging control flags with control flags2
        if "51 Control flags2" in accessory.columns and "51 Control flags" in accessory.columns:
            accessory["51 Control flags2"] = accessory["51 Control flags2"].fillna(accessory["51 Control flags"])
            accessory = accessory.drop("51 Control flags", axis="columns")
        elif "51 Control flags" in accessory.columns:
            accessory = accessory.rename(columns={"51 Control flags":"51 Control flags2"})
            
        columns = accessory.columns
        new_columns = ["51 Control flags2" if i == "51 Control flags" else i for i in columns]
        new_columns = [column[column.find(" ")+1:] for column in new_columns]
        
        ##### sql column headers can only be letters or underscores, so replace invalid characters with underscores
        ##### also cant start with a number, so preface those columns with an underscore
        for i, col in enumerate(new_columns):
            new_columns[i] = new_columns[i].replace("-", "_")
            new_columns[i] = new_columns[i].replace(" ", "_")
            new_columns[i] = new_columns[i].replace("%", "")
            if col[0].isnumeric():
                new_columns[i] = "_"+col
            
        accessory = accessory.rename(columns=dict(zip(columns, new_columns)))
        ##### merge the gps time columns into one time column
        gps_date_cols = ["GPSYYYY", "GPSMonth", "GPSDay"]
        gps_time_cols = ["GPSHour", "GPSMin", "GPSSec"]
        accessory = self.__merge_date_cols(accessory, "gps_time", gps_date_cols, gps_time_cols)
        accessory = accessory.drop(gps_date_cols + gps_time_cols, axis="columns")
        
        pc_date_cols = ["TimeYYYY", "Month", "Day"]
        pc_time_cols = ["Hour", "Minute", "Second"]
        accessory = self.__merge_date_cols(accessory, "cpu_time", pc_date_cols, pc_time_cols, "milliseconds", "Timezone")
        accessory = accessory.drop(pc_date_cols + pc_time_cols + ["milliseconds", "Timezone"], axis="columns")
        
        accessory = self.reset_index(accessory, "pc_time_end_measurement")
        
        return accessory
    
    
    def calculate_timezone(self, timezone):
        ##### reads timezone data from ini file and converts utc time to tzaware timestamp, stored as string
        if timezone[0] != "+" and timezone[0] != "-":
            timezone = "+"+timezone
        timezone_components = timezone.split(":")
        
        mins = 0
        if len(timezone_components) == 2:
            mins = abs(int(timezone_components[0]))*60 + abs(int(timezone_components[1]))
        else:
            mins = abs(int(timezone))*60
        
        if timezone[0] == "-":
            mins = -mins
            
        timedelta = pd.Timedelta(mins, "minutes")
        return timezone, timedelta
    
    
    def __merge_date_cols(self, df, name, date_cols, time_cols, millisecond="", timezone=""):
        df[date_cols] = df[date_cols].astype(int).astype(str)
        df[time_cols] = df[time_cols].astype(int).astype(str)
        
        ##### zero pad dates and times
        for date_col in date_cols:
            df.loc[df[date_col].str.len() == 1, date_col] = "0"+df.loc[df[date_col].str.len() == 1, date_col]
        for time_col in time_cols:
            df.loc[df[time_col].str.len() == 1, time_col] = "0"+df.loc[df[time_col].str.len() == 1, time_col]
        
        timezones = milliseconds = []
        if millisecond == "":
            milliseconds = ""
        else:
            milliseconds = df[millisecond].astype(int).astype(str)
        
        if timezone == "":
            timezones = ""
        else:
            timezones =  df[timezone].astype(int).astype(str)
        
        
        ##### these lines are the (speed) problem \/
        dates = df[date_cols].agg("-".join, axis=1)
        times = df[time_cols].agg(":".join, axis=1)
        
        ##### to avoid trailing .
        if type(milliseconds) is str:
            df[name] = dates + " " + times + milliseconds + timezones
        else:
            df[name] = dates + " " + times + "." + milliseconds + timezones
        
        return df
    
    def __round_times(self, df:pd.DataFrame, frequency:str) -> pd.DataFrame:
        """rounds the index of a dataframe to the nearest value, as passed
        params:
            df: dataframe to be rounded
            frequency: what to round the database to ('min', 'hour' etc)
            https://pandas.pydata.org/docs/user_guide/timeseries.html#timeseries-offset-aliases
        returns:
            df: the rounded dataframe
        """
        df.index = df.index.round(frequency)
        return df
    
    def match_sample_ids(self, source:pd.DataFrame, to_be_matched:pd.DataFrame, column:str) -> pd.DataFrame:
        """matches the sample ids from a dataframe with sample ids(or other columns) to one without, where they share a timestamp
        params:
            source: dataframe that already has ids
            to_be_matched: dataframe that will take the id from source where timestamps match
            column: the column that will be transferred from source, usually sample_id
        returns: to_be_matched: the dataframe with the matched ids added
        """
        
        # TODO: to make faster: subdataframes and pd.merge
        # probably dosent matter, already very fast, 0.02s for 2months of data
        to_be_matched[column] = np.nan
        to_be_matched[column] = to_be_matched[column].fillna(source[column])
        return to_be_matched
    
    def listify(self, df):
        """converts a dataframe with many columns into a series with an array representing each row"""
        return pd.Series([np.array(x[1]) for x in df.iterrows()], df.index)
    
    def reset_index(self, df, column_name):
        """copies the index into a column with the name column_name and sets the index to integers
        params:
            df: dataframe that will be changed
            column_name: new name for the old index
        returns:
            df: dataframe with the new column
        
        this method can be replaced with pandas' reset_index names= parameter, but that's new in pandas v1.4,
        so ive added this for backwards compatibility
        """
        df = df.reset_index()
        columns = df.columns.to_list()
        columns[0]=column_name
        df.columns = columns
        return df
        


# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""

import numpy as np
import pandas as pd
import scipy as sp
import matplotlib.pyplot as plt
import matplotlib


from hsr1.plots import graphUtils as utils
from hsr1.utils.spectrum import spectrumUtils as SpectrumUtils



class DailyDipsSummary:
    def __init__(self, timezone=None, reference_lines=None, reference_labels=None):
        self.reference_lines = reference_lines
        self.reference_labels = reference_labels
        self.timezone = timezone
        
    def plot_daily_dips_hist(self, global_spectrum, n=15, cutoff_wavelength=1000):
        dates = global_spectrum["pc_time_end_measurement"]
        
        global_spectrum = global_spectrum["global_spectrum"].values
        
        nlargest, _ = SpectrumUtils.find_nlargest(global_spectrum, n, ["wavelengths", "prominences"], cutoff_wavelength, fill_nan=True)
        
        if len(nlargest) == 0:
            raise ValueError("no dips found")
        
        not_nan = np.logical_not(np.isnan(nlargest))
        nlargest = nlargest[not_nan]
        
        ##### generate a list of all the days in the dataset
        dates = pd.to_datetime(dates.dt.date)#.dt.tz_localize(self.timezone)
        dates_int = dates.astype("int64")
        all_dates = []
        [all_dates.append([x]*n) for x in dates_int]
        all_dates = np.array(all_dates).flatten()
        all_dates = all_dates[not_nan]
        
        
        fig = plt.figure(layout="constrained", figsize=(11.7, 8.3))
        axes = fig.subplots(1)
        
        ##### weight each day to the number of readings in the day
        weights = []
        for date in dates.unique():
            readings = dates.loc[dates==date]
            weights += [1440/len(readings)]*(len(readings)*n)
        weights = np.array(weights)
        weights = weights[not_nan]
        
        total_num_days = len(pd.date_range(min(dates), max(dates)))
        bins = (total_num_days, int((cutoff_wavelength-300)/1))
        
        ##### matplotlib dosent like when the data array is all the same and also large.
        #####   changing the all_dates variable to be 1 solves this issue.
        #####   xtick is taken from dates so is still valid
        if all_dates[0] == all_dates[-1]:
            all_dates = np.ones(len(all_dates))
            all_dates[-1] += 1
            
        
        hist, *_ = np.histogram2d(all_dates, nlargest, bins=bins, weights=weights, range=[[all_dates[0], all_dates[-1]], [300, bins[1]+300]])
        hist = hist.T
        hist = np.flip(hist, 0)
        hist[hist == 0.0] = np.nan
        
        extent = (all_dates[0], all_dates[-1], 300, bins[1]+300)
        image = axes.imshow(hist, cmap="jet", vmin=1, aspect="auto", extent=extent, interpolation="none")        
        
        axes.set_ylim(300, cutoff_wavelength)
        
        xticks, xlabels = utils.calculate_date_labels(dates, axes.set_xlim())
        axes.set_xticks(xticks, xlabels)
        
        # cbar = fig.colorbar(hist[3], ax=axes, aspect=25, pad=0.01)
        cbar = fig.colorbar(image, ax=axes, aspect=25, pad=0.01)
        
        utils.plot_reference_lines_and_labels(axes, all_dates[0], self.reference_lines, self.reference_labels)
        
        
        axes.set_ylabel("Wavelength, \nDip Locations from SMARTS")
        
        return fig
    
    
    def plot_daily_peaks_summary(self, global_spectrum, n, cutoff_wavelength):
        fig = self.plot_daily_dips_hist(global_spectrum, n, cutoff_wavelength)
        
        return fig


# -*- coding: utf-8 -*-
"""
Created on Wed Feb 28 16:12:21 2024

@author: albie
"""

from hsr1.plots.clearnessDensityGraph import ClearnessDensityGraph
from hsr1.plots.dailyDipsSummary import DailyDipsSummary
from hsr1.plots.dailyHists import DailyHists
from hsr1.plots.dailyPlots import DailyPlots
from hsr1.plots.elvAziGraph import ElvAziGraph
from hsr1.plots.latLonGraph import LatLonGraph
from hsr1.plots.linearDipsGraph import LinearDipsGraph
from hsr1.plots.linearTimeGraph import LinearTimeGraph
from hsr1.plots.timeDayGraph import TimeDayGraph
from hsr1.plots.spectrumGraph import SpectrumGraph

import hsr1.plots.flagData as flagData
import hsr1.plots.graphUtils as graphUtils



# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""
import time
import datetime

import pandas as pd
import numpy as np
import matplotlib
import ephem


def calculate_date_labels(time_series:pd.Series, xlims, target_num_ticks:int=10):
    """evenly arranges date labels along the bottom of a daily plot.
    only works with discrete days, not continuous time data
    params: 
        time_series: series conatining all the dates
        xlims: limits of the x axis [smallest, largest]
        num_ticks: number of ticks to be displayed
    """
    time_series = time_series.astype(str)
    times = time_series.str.slice(stop=10).unique()
    times.sort()
    dates = pd.date_range(times[0], times[-1]).strftime("%d-%b")
    # dates = pd.DatetimeIndex(times).strftime("%d-%b")
    
    num_days = len(dates)
    
    ##### locations that are in the middle of a day
    start_of_day = np.arange(xlims[0], xlims[-1], (xlims[-1]-xlims[0])/num_days)
    valid_tick_spots = start_of_day + (((xlims[-1]-xlims[0])/num_days)/2)
    
    
    period = ((num_days-1)//(target_num_ticks))+1
    
    
    ##### finds a consistent period that will result in as close to target_num_ticks as possible
    xtick_indexes = np.arange(0, num_days, period).astype(int)
        
    xticks = [valid_tick_spots[i] for i in xtick_indexes]
    labels = [dates[i] for i in xtick_indexes]
    
    return xticks, labels


def cbar(axes, fig, max_integral:int, tick_period=100, aspect=25):
    """plots a colorbar and generates labels for it
    params:
        axes: axes or list of axes that the colorbar will steal space from
        fig: figure all the graphs are plotted on
        max_integral: the top of the colorbar
        tick_period: distance between the ticks
        aspect: aspect ratio, 1=square
    """
    cbar = fig.colorbar(matplotlib.cm.ScalarMappable(cmap="jet"), ax=axes, aspect=25, pad=0.01)
    
    yticks = np.arange(0, max_integral+1, tick_period)
    ytick_location = yticks/max_integral
    ytick_location[ytick_location == np.inf] = 0
    cbar.ax.set_yticks(ytick_location, yticks.astype(int))


def calculate_sunrise_sunset(data:pd.DataFrame, timezone:pd.Timedelta):
    """
    params:
        data: dataframe with timestamps and locations in
        timezone: timezone of the data, pd.Timedelta
    returns:
        sunrise_set_df: dataframe containing times of the sunrise and sunset
    """
    df = data.copy()
    df = df.dropna()
    
    df["pc_time_end_measurement"] = pd.DatetimeIndex(df["pc_time_end_measurement"].dt.tz_convert(None))
    
    dates = df["pc_time_end_measurement"].dt.date.unique()
    
    ##### set middle of data, rather than the start
    new_df = pd.DataFrame(index=pd.DatetimeIndex(dates) + pd.Timedelta(12, "hours"))
    
    df = df.sort_values("pc_time_end_measurement")
    new_df = new_df.sort_index()
    
    new_df = pd.merge_asof(new_df, df, left_index=True, right_on="pc_time_end_measurement", direction="nearest")
    new_df[["gps_longitude", "gps_latitude", "gps_altitude"]] = new_df[["gps_longitude", "gps_latitude", "gps_altitude"]].astype(float)
    new_df = new_df.ffill()
    
    
    geopoint = new_df[["gps_longitude", "gps_latitude", "gps_altitude"]].values
    timestamp = pd.to_datetime(new_df.index)
    
    
    sunrise_set_df = pd.DataFrame(index=timestamp)
    
    sun = ephem.Sun()
    
    sunrises = []
    sunsets = []
    for i in range(len(timestamp)):
        instrument = ephem.Observer()
        instrument.date = timestamp[i]
        instrument.lat = str(geopoint[i][1])
        instrument.lon = str(geopoint[i][0])
        instrument.elevation = geopoint[i][2]
        
        try:
            sunrise = instrument.next_rising(sun)
            sunrises.append(sunrise.datetime())
        except ephem.AlwaysUpError:
            sunrises.append(None)
        except ephem.NeverUpError:
            sunrises.append(None)
        
        try:
            sunset = instrument.next_setting(sun)
            sunsets.append(sunset.datetime())
        except ephem.AlwaysUpError:
            sunsets.append(None)
        except ephem.NeverUpError:
            sunsets.append(None)
    
    ##### convert back to local time
    sunrise_set_df["sunrise"] = np.array(sunrises)+timezone
    sunrise_set_df["sunset"] = np.array(sunsets)+timezone
    
    return sunrise_set_df


    
def plot_reference_lines_and_labels(axes, left_edge=0, reference_lines=None, reference_labels=None):
    if reference_lines is None:
        reference_lines = [374, 384, 393, 430, 487, 517, 590, 657, 688, 719, 761, 817, 900, 936, 976]
    if reference_labels is None:
        reference_labels = ["Fe", "Ca+", "Ca+", "Ca/Fe", "H-B", "Mg", "Na", "H-A", "O2", "H2O", "O2", "O2", "", "", ""]
    elif reference_labels == []:
        reference_labels = [""]*len(reference_lines)
    
    
    fancy_strings = []
    ##### add dashes if not epmpty string
    for j in range(len(reference_labels)):
        reference_labels[j] = reference_labels[j] +" - " if reference_labels[j] != "" else ""
        fancy_strings.append(reference_labels[j]+str(reference_lines[j]))
    
    ##### plus 0.5 so lines are in the middle of the pixel, not the bottom
    reference_lines = np.array(reference_lines)+0.5
    for i in range(len(reference_lines)):
        axes.axhline(y=reference_lines[i], color="lightgray", linewidth=0.5, zorder=1.5)

    axes.set_yticks(reference_lines, fancy_strings)
    axes.tick_params(axis='y', which='major', labelsize=10)

def superscript(number):
    if type(number) != int:
        raise TypeError("number to be superscripted must be an int")
    superscript_chars = "⁰¹²³⁴⁵⁶⁷⁸⁹"
    result = ""
    if number < 0:
        result = "⁻"
        number = abs(number)
    for num in str(number):
        result += superscript_chars[int(num)]
    
    return result



def generate_metadata_string(deployment_metadata, data=None):
    #### make a dict of name: value pairs
    metadata = {}
    
    metadata["lat"] = deployment_metadata['default_latitude'][0]
    metadata["lon"] = deployment_metadata['default_longitude'][0]
    
    if data is not None:
        metadata["period"] = str(data.iloc[0]["pc_time_end_measurement"].date()) + " - " + str(data.iloc[len(data)-1]["pc_time_end_measurement"].date())
        metadata["total readings"] = str(len(data))
    
    ##### convert dict to string
    output = ""
    for line in metadata.keys():
        if metadata[line] != "''":
            output += line+": "+metadata[line]+", "
    
    ##### crop trailing ", "
    return output[:-2] + "\n"

def total_seconds(timestamp):
    return (timestamp - datetime.datetime(1970, 1, 1, )).total_seconds()

def make_full_title(title):
    return title.replace("<", "_").replace(">", "_").replace(":", "_").replace("/", "_").replace("\\", "_").replace("|", "_").replace("?", "_").replace("*", "_").replace("\"", "_").replace("\n", " ").replace("\"", "'").replace("|", "_")


def numpy_string_slice(a, start, end):
    b = a.view((str,1)).reshape(len(a),-1)[:,start:end]
    return np.fromstring(b.tostring(),dtype=(str,end-start))



# -*- coding: utf-8 -*-
"""
Created on Wed Feb 28 16:07:38 2024

@author: albie
"""
from hsr1.db.preCalculations import PreCalculations
from hsr1.db.serialisation import Serialisation
from hsr1.db.sqliteDBLoad import SqliteDBLoad
from hsr1.db.sqliteDBStore import SqliteDBStore



# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""
import os
import datetime as dt
import time
import importlib_resources

import numpy as np
import pandas as pd
import matplotlib.pyplot as pl
import matplotlib.dates as mdates
import ephem
import scipy.interpolate as interpolate
import zipfile



def calc_direct_normal_spectrum(global_spectrum, diffuse_spectrum, sza):
    """calculates the direct normal spectrum from the diffuse, global and zenith angle
    params:
        global_spectrum, diffuse_spectrum: either pandas Series of numpy arrays, each array representing one spectral reading
                                                  or dataframe with one row per wavelength reading
        sza: pandas Series one float(radians) per reading
    """
    direct_horizontal_spectrum = global_spectrum-diffuse_spectrum
    direct_normal_spectrum = direct_horizontal_spectrum/np.cos(sza)
    
    return direct_normal_spectrum
    

def calc_sun_zenith(ts, lat, lon):
    """calculates the solar zenith angle at a specific time and location
    quite slow as this has to be run in a for loop, sg2's calculation is much faster
    params:
        ts: timestamp that you want to measure, string yyyy-mm-dd hh:mm:ss
        lat, lon: latitude and longitude of the location
    
    returns:
        tuple of zenith angle and azimuth, in radians
    """
    obs = ephem.Observer()
    sun = ephem.Sun()

    obs.date = ts
    obs.lat, obs.lon = str(lat), str(lon)

    sun.compute(obs)

    return np.pi/2 - sun.alt, sun.az

def calc_rayleigh(wl, pressure = 1013.25):
    """calculates the rayleigh scattering
    params:
        wl: the wavelength(s) to calculate the scattering over.
            this can be a scalar, numpy array or pandas series
        pressure: the pressure in hPa
    return is in the same datatype as the wl parameter
    """
    a1 = 117.25942 # model Rayleigh scattering 
    a2 = -1.3215
    a3 = 0.00032073
    a4 = -0.000076842
    p0 = 1013.25 
    wl  = wl*1e-3 # convert wl from nano to micrometres (for Rayleigh formula)    
    tau_r = (pressure/p0)*(1/(a1*wl**4 + a2*wl**2 + a3 + a4/wl**2))
    return tau_r

def calc_air_mass(Sza, pressure=1013.25):
    """calculates the airmass
    params:
        Sza: the solar zenith angle in radians.
            this can be a scalar, numpy array or pandas series.
        pressure: the pressure in hPa
    return is in the same datatype as the wl parameter
    """
    a = 0.50572 
    b = 96.07995
    c = 1.6364
    C = np.cos(Sza)
    
    mu = C + a*(b - np.degrees(Sza))**(-c) # atm. air mass (note 1/m in Wood et. al 2019)
    return pressure / 1013.25 / mu

def calc_aot_direct(ed, eds, sza, e_solar=None, sed=None, aod_type=["total_od", "aod_microtops", "aod_wood_2017"], et_wavelengths=None):
    """calculates the atmospheric optical depth in several different ways
    
    params:
        ed: the global spectral irradiance. columns=wavelength, index=time, timestamp in utc
        eds: the diffuse spectral irradiance. columns=wavelength, index=time, timestamp in utc
        sza: the solar zenith angle in radians. column = "sza"
        e_solar: the extraterrestrial solar spectrum. columns=wavelength, index=time, timestamp in utc
        sed: sun earth distance in AU, pandas Series. if None, calculated from time. this is quite slow so this parameter
            can speed it up if you are already loading from the database
        aod_type: the desired outputs string or list of strings
        
    
    returns:
        aod_data: dataframe containing all the requested aod_types
        to convert a column of numpy arrays into a dataframe of wavelengths against time:
            pd.DataFrame(np.stack(data["spectrum_column"].values), columns = np.arange(300, 1101, 1), index=data["pc_time_end_measurement"])
    
    definition of each aod type:
        tau_t/total_od: the total optical depth from all sources
        tau_a/microtops_aod: the aerosol optical depth. this is the optical depth with rayleigh scattering removed
        tau_corr/aod_wood_2017: the aerosol optical depth with an empirical correction applied to it
    """
    if e_solar is None:
        e_solar = load_et_spectrum(wavelengths=et_wavelengths)
    
    
    edd = (ed - eds)
    
    edni = edd.divide(np.cos(sza.values.astype(float)), axis="index")
    
    sun = ephem.Sun()
    
    tau_r = calc_rayleigh(ed.columns.astype(float))
    tau_t = np.nan*np.ones([len(edni),len(edni.T)]) # total OT 
    tau_a = np.nan*np.ones([len(edni),len(edni.T)]) # aerosol OT 
    tau_corr = np.nan*np.ones([len(edni),len(edni.T)]) # corrected OT
    
    wl_e = np.array([440, 500, 675, 870, 1020]) #  empirical correction coefficients (Wood 2017)
    offset_am = np.array([0.0097, 0.0177, -0.0033, -0.0067, -0.0117])
    offset_wl = np.array([0.0244, 0.0260, 0.0182, 0.0124, 0.0457])
    slope_wl = np.array([1.2701, 1.2893, 1.3549, 1.4522, 1.5237])
    
    wl = ed.columns.astype(float)
    offset_am = interpolate.interp1d(wl_e, offset_am, kind = 'linear', axis = 0, fill_value = "extrapolate") # interpolate to wl range of hsr (piecewise linear)
    offset_wl = interpolate.interp1d(wl_e, offset_wl, kind = 'linear', axis = 0, fill_value = "extrapolate")(wl) # linearly extrapolated outside wl range
    slope_wl = interpolate.interp1d(wl_e, slope_wl, kind = 'linear', axis = 0, fill_value = "extrapolate")(wl)
    
    daytime = sza.values < np.radians(90)
    daytime = daytime[:, 0]
    
    
    
    if sed is None:
        sed = np.full(len(edni), np.nan)
        for i in range(len(edni)):
            if sza.iloc[i].values < np.radians(90):
                sun.compute(edni.index[i])
                sed[i] = sun.earth_distance
        sed = pd.Series(sed)
    
    ##### vectorised calculations
    airmasses = calc_air_mass(sza).values[:, 0]
    
    e_toa = pd.DataFrame((e_solar.values/np.array([sed.values]).T**2), columns=e_solar.columns, index=ed.index)
    
    airmasses = airmasses.T[:, np.newaxis]
    offsets = offset_am(airmasses)
    
    dnitoa =edni.divide(e_toa).astype(float)
    tau_t = -1/airmasses*np.log(dnitoa)
    tau_a = tau_t-tau_r.values
    tau_corr = (tau_a - offsets - offset_wl) *slope_wl
    
    tau_t[np.logical_not(daytime)] = np.nan
    tau_a[np.logical_not(daytime)] = np.nan
    tau_corr[np.logical_not(daytime)] = np.nan
    
    aod_data = pd.DataFrame()
    aod_data["pc_time_end_measurement"] = ed.index
    
    if isinstance(aod_type, str):
        aod_type = [aod_type]
    
    if "total_od" in aod_type:
        aod_data["total_od"] = list(tau_t.values)
    if "aod_microtops" in aod_type:
        aod_data["aod_microtops"] = list(tau_a.values)
    if "aod_wood_2017" in aod_type:
        aod_data["aod_wood_2017"] = list(tau_corr.values)
    
    return aod_data


def calc_cimel_band_aot_direct(Ed, Eds, Sza, E_solar, sed, aod_type=["total_od", "aod_microtops", "aod_wood_2017"]):
    """calculates the AOD values at several wavelengths, that are also measured by cimel instruments
    
    params:
        ed: the global spectral irradiance. columns=wavelength, index=time, timestamp in utc
        eds: the diffuse spectral irradiance. columns=wavelength, index=time, timestamp in utc
        sza: the solar zenith angle in radians. column = "sza", index=time, timestamp in utc
        e_solar: the extraterrestrial solar spectrum. columns=wavelength, index=time, timestamp in utc
        sed: sun earth distance in AU. if None, calculated from time. this is quite slow so this parameter
            can speed it up if you are already loading from the database
        aod_type: the desired outputs string or list of strings
        
    
    returns:
        aod_data: dataframe containing all the requested channels
        to convert a column of numpy arrays into a dataframe of wavelengths against time:
            pd.DataFrame(np.stack(data["spectrum_column"].values), columns = np.arange(300, 1101, 1), index=data["pc_time_end_measurement"])
    
    definition of each aod type:
        tau_t/total_od: the total optical depth from all sources
        tau_a/microtops_aod: the aerosol optical depth. this is the optical depth with rayleigh scattering removed
        tau_corr/aod_wood_2017: the aerosol optical depth with an empirical correction applied to it
    """
    
    cimel_band = [380, 440, 500, 675, 870, 1020]
    #Ed_cb = Ed[list(map(str,cimel_band))]
    #Eds_cb=Eds[list(map(str,cimel_band))]
    #solar_cb=E_solar[list(map(str,cimel_band))]
    Ed_cb = Ed[cimel_band]
    Eds_cb = Eds[cimel_band]
    solar_cb = E_solar[cimel_band]
    
    Ed_cb_i = []
    Eds_cb_i = []
    solar_cb_i = []
    for i in range(len(cimel_band)):
        idxrange = list(range(cimel_band[i]-5,cimel_band[i]+6))
        Ed_cb_i.append(Ed[idxrange].mean(axis=1))
        Eds_cb_i.append(Eds[idxrange].mean(axis=1))
        solar_cb_i.append(E_solar[idxrange].mean(axis=1))
    Ed_cb = pd.concat(Ed_cb_i, axis = 1)
    Eds_cb = pd.concat(Eds_cb_i, axis = 1)
    solar_cb = pd.concat(solar_cb_i, axis = 1)
    Ed_cb.columns = cimel_band
    Eds_cb.columns = cimel_band
    solar_cb.columns = cimel_band
         
    return calc_aot_direct(Ed_cb, Eds_cb, Sza, solar_cb, sed, aod_type)
    

def calc_aod_from_df(data, cimel=False, aod_type=["total_od", "aod_microtops", "aod_wood_2017"], wavelengths=None):
    """coerces data from the format returned from the database to the format for the calculation, and back,
    also loads the reference solar spectrum
    params:
        data: dataframe of all the relevant data, columns = ["pc_time_end_measurement", 
                                                             "global_spectrum", "diffuse_spectrum",
                                                             "sza", "sed"]
        cimel: if True, calculates the aod only at certain wavelengths that cimel instruments measure at
        aod_type: the desired outputs string or list of strings
        wavelengths: filters the wavlengths that are used for the caluclation, and that are returned
    """
    utctime = pd.to_datetime(data["pc_time_end_measurement"]).dt.tz_localize(None)
    
    ed = pd.DataFrame(np.stack(data["global_spectrum"].values), columns = np.arange(300, 1101, 1), index=utctime)
    eds = pd.DataFrame(np.stack(data["diffuse_spectrum"].values), columns = np.arange(300, 1101, 1), index=utctime)
    sza = pd.DataFrame(data["sza"])
    
    if wavelengths is None:
        wavelengths = np.arange(300, 1101, 1)
    else:
        ed = ed[wavelengths]
        eds = eds[wavelengths]
    
    e_solar_df = load_et_spectrum(wavelengths=wavelengths)
    
    aod_data = None
    if cimel:
        aod_data = calc_cimel_band_aot_direct(ed, eds, sza, e_solar_df, data["sed"], aod_type)
    else:
        aod_data = calc_aot_direct(ed, eds, sza, e_solar_df, data["sed"], aod_type)
    
    return aod_data

def load_et_spectrum(filepath=None, wavelengths=None):
    """loads a .txt file containing a reference extraterrestrial solar spectrum, and applies some smoothing
    params:
        filepath: filepath to the file where the spectrum is stored. if None, 
            a default spectrum that comes with the library is used
        wavelengths: which wavelengths to use from the reference file
    returns a dataframe with one row which contains the reference spectrum
    """
    
    if wavelengths is None:
        wavelengths = np.arange(300, 1101)
    
    if filepath is None:
        res = importlib_resources.files("hsr1.data").joinpath("SolarSpectrum.txt")
        file = importlib_resources.as_file(res)
        with file as f:
            filepath = f
    
    e_solar = pd.read_csv(filepath, skiprows=1, delimiter='\t', index_col=0)
    smoothed_e_solar = e_solar.rolling(3).mean().T
    smoothed_e_solar = smoothed_e_solar.fillna(e_solar.T)
    
    e_solar_df = pd.DataFrame(columns=wavelengths)
    e_solar_df.loc[0, :] = smoothed_e_solar[wavelengths].values
    
    return e_solar_df


def calculate_clearsky_wood(data, global_spectrum=None,
                            diffuse_spectrum=None,
                            column="total_od", 
                            absolute_filter:float=2, 
                            relative_filter:float=0.05,
                            relative_time_period:str="10min"):
    """calculates which readings are cloudfree
    
    params:
        data: dataframe with at least columns 
            ["pc_time_end_measurement", "global_spectrum", "diffuse_spectrum", "sza", "sed"]
        column: which aod column to use, one of "total_od", "aod_microtops", "aod_wood_2017"
        absolute_filter: any readings higher than this will be filtered out
        relative_filter: any readings that are more than this value more or less than any
            other within the time window will be filtered out
        relative_time_period: the time period over which the relative filtering is done.
            this is the total time period that the filtering is done by, so if you want
            5 mins either side, pass 10min
            format is a string that will be passed to pd.Timedelta. documentation:
            https://pandas.pydata.org/docs/reference/api/pandas.Timedelta.html
    
    returns a numpy array the same length as input dataframe with ones and zeros, 1=clearsky 0=cloud
    """
    aod_data = None
    if global_spectrum is None and diffuse_spectrum is None:
        aod_data = calc_aod_from_df(data, aod_type=column, wavelengths=[500])
    else:
        aod_data = calc_aot_direct(pd.DataFrame(global_spectrum[500]), pd.DataFrame(diffuse_spectrum[500]), pd.DataFrame(data["sza"]), et_wavelengths=[500])
    nm500 = np.stack(aod_data[column].values)[:, 0]
    nm500df = pd.DataFrame(index=pd.DatetimeIndex(data["pc_time_end_measurement"]))
    nm500df["base"] = nm500
    nm500df["max"] = nm500df.rolling(pd.Timedelta(relative_time_period), center=True)["base"].max()-nm500df["base"]
    nm500df["min"] = nm500df.rolling(pd.Timedelta(relative_time_period), center=True)["base"].min()-nm500df["base"]
    nm500df["max"] = np.abs(nm500df["max"])
    nm500df["min"] = np.abs(nm500df["min"])
    nm500df["abs"] = np.logical_and(nm500df["max"] < relative_filter, nm500df["min"] < relative_filter)
    nm500df_abs = np.logical_and(nm500df["max"] < relative_filter, nm500df["min"] < relative_filter)
    
    clearsky_filter = np.logical_and(nm500df_abs.values, nm500 < absolute_filter)
    return clearsky_filter


def calculate_clearsky_filter(data:pd.DataFrame, global_spectrum=None, diffuse_spectrum=None, method:str="wood", kwargs:dict={}):
    """method that calls the appropriate clearsky function
    
    params:
        data: the data that is used for the clearsky calculation
        method: the method that iwll be used to calculate which readings are clearsky readings
        kwargs: the keyword arguments to pass to clearsky_filter
    returns a numpy array the same length as input dataframe with ones and zeros, 1=clearsky 0=cloud
    """
    if method is None:
        return np.ones(len(data)).astype(bool)
    if method == "wood":
        return calculate_clearsky_wood(data, global_spectrum, diffuse_spectrum, **kwargs)
    
    print("filtering method not recognised, not filtering")
    return np.ones(len(data)).astype(bool)




def Get_hsr_Dates(hsr_path, start_date, end_date):
# Get a list of hsr datafiles in the root folder

    folderlist = None
    if hsr_path == "":
        folderlist = os.listdir()
    else:
        folderlist = os.listdir(hsr_path)
    
    datelist = []
    for i, name in enumerate(folderlist):
        try:
            res = bool(dt.datetime.strptime(name[:10], '%Y-%m-%d'))
            datelist.append(name[:10])
        except ValueError:
            res = False
            
    date_dict = dict.fromkeys(datelist, True)     # removes any duplicate dates
    try: 
        # check dates are in range
            for i, name in enumerate(date_dict):
                if len(start_date):
                    if dt.datetime.strptime(name, '%Y-%m-%d') < dt.datetime.strptime(start_date[:10], '%Y-%m-%d'):   
                        date_dict[name] = False
                if len(end_date):
                    if dt.datetime.strptime(name, '%Y-%m-%d') > dt.datetime.strptime(end_date[:10], '%Y-%m-%d'):   
                        date_dict[name] = False
               
    except:
        print('Error checking dates')
        print("start_date: "+str(start_date))
        print("end_date: "+str(end_date))
    
    #print (date_dict)
    hsr_dates  = [k for k, v in date_dict.items() if v == True]
    return hsr_dates






# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""
import datetime
import os, io
import numpy as np
import pandas as pd
import matplotlib.pyplot as pl
import matplotlib.dates as mdates
import datetime as dt
import zipfile 

import ephem

def calc_sun_zenith_azimuth(ts, lat, lon):
    obs = ephem.Observer()
    sun = ephem.Sun()

    obs.date = ts
    obs.lat, obs.lon = str(lat), str(lon)

    sun.compute(obs)

    return 90 - (sun.alt * 180. / np.pi)#, sun.az * 180. / np.pi

def open_gps_file(hsr_path, hsr_date):
    # hsr_path is root folder for the dataset, hsr date is text representation of the date
    filename = os.path.join(hsr_path, hsr_date, '.zip')
    GPS = []
    if os.path.isfile(filename):
        try:    # try the quick read, works for uninterrupted files
            archive = zipfile.ZipFile(filename, 'r')
            ### Import hsr GPS from daily folder
            #GPS = pd.read_csv(archive.open( 'GPS.txt'), skiprows=1, delimiter='\t', parse_dates=['PC Time','GPS Time'], index_col=0, converters = {'Status': str})
            with archive.open('GPS.txt') as datafile:
                GPS = read_by_line(datafile, hsr_date)
            
        except:     # try a line-by-line read
            print ("Error reading zip, trying line-by-line " + filename)

    elif os.path.isdir(os.path.join(hsr_path, hsr_date)):
        filename = os.path.join(hsr_path, hsr_date, 'GPS.txt')
        try:    # try the quick read, works for uninterrupted files
        
            #GPS = pd.read_csv(filename, skiprows=1, delimiter='\t', parse_dates=['PC Time','GPS Time'], index_col=0, converters = {'Status': str})
            with open(filename) as datafile:
                GPS = read_by_line(datafile, hsr_date)
        except:     # try a line-by-line read
            print ("Error reading, trying line-by-line  " + filename)
    
    return GPS

def read_by_line(datafile, hsr_date):
    # print('read-by-line ' + hsr_date + ' ' + os.path.basename(datafile.name))
    row_list=[]        # list which stores date time row objects
    dt_list = []        # list which stores the actual date-times
    element_list=[] # list which stores the elements of each row
    column_list = np.arange(300,1101)    # list which stores the column headings, default is for Raw files
    for i, row in enumerate(datafile):
        dict1 = {}
        if row[:10] == hsr_date: # tests for line which contains radiance entries (first 10 elements are date)
            date_time_string=row[0:19]
            if not date_time_string:  # break condition for empty string
                break
            element_list = row.strip().split("\t")[1:]      # strip gets rid of line ends in last item
            if len(column_list) <= len(element_list):
                for count, value in enumerate(column_list):
                    try:
                        # dict1.update({value:float(element_list[count])}) ##### <- previous line, access by key is a bit faster
                        dict1[value] = float(element_list[count])    # convert to a number if it will
                    except ValueError:
                        dict1[value] = element_list[count]       # otherwise treat as a string
                row_list.append(dict1)
                dt_list.append(dt.datetime.strptime(date_time_string,'%Y-%m-%d %H:%M:%S'))
        elif  'Time' in row[:10]:
            column_list = row.strip().split("\t")[1:]
    # now turn it into a dataframe
    df = pd.DataFrame(row_list)
    if len(row_list) == 0:
        return pd.DataFrame()
    df.columns=column_list
    df.index = dt_list
    return  df 
    
def open_hsr_file(hsr_path, hsr_date, hsr_file, Raw=False):
    # hsr_path is root folder for the dataset, hsr_date is text representation of the date
    filename = os.path.join(hsr_path, hsr_date + '.zip')
    if hsr_file == 'GPS.txt':
        time_cols = ['PC Time', 'GPS time']
    else:
        time_cols = ['Time']
        
    hsr_df = []
    if os.path.isfile(filename):    # for data in zipfiles
        try:    # try the quick read, works for uninterrupted files
            archive = zipfile.ZipFile(filename, 'r')
            ### Import hsr file from daily folder
            if Raw==True:
                hsr_df = pd.read_csv(archive.open( hsr_file), skiprows=1, delimiter='\t', index_col=0, header=None)                
                hsr_df.columns = np.arange(300, 1101)
            else:
                hsr_df = pd.read_csv(archive.open( hsr_file), skiprows=2, delimiter='\t',parse_dates=time_cols,  index_col=0)
            hsr_df.index = pd.to_datetime(hsr_df.index)
        except KeyError:
            return hsr_df
        except:     # try a line-by-line read
            # print ("Error reading zip, trying line-by-line " + filename)
            with io.TextIOWrapper(archive.open(hsr_file), encoding="utf-8") as datafile:
                hsr_df = read_by_line(datafile, hsr_date)

    elif os.path.isdir(os.path.join(hsr_path, hsr_date)):     # for data expanded into dated folders
        filename = os.path.join(hsr_path, hsr_date, hsr_file)
        try:    # try the quick read, works for uninterrupted files
            if Raw==True:
                hsr_df = pd.read_csv(filename, skiprows=1, delimiter='\t', index_col=0, header=None)                
                hsr_df.columns = np.arange(300, 1101)
            else:        
                hsr_df = pd.read_csv(filename, skiprows=2, delimiter='\t', parse_dates=time_cols, index_col=0)
            hsr_df.index = pd.to_datetime(hsr_df.index)
        except KeyError:
            return hsr_df
        except:     # try a line-by-line read
            # print ("Error reading, trying line-by-line  " + filename)
            try: 
                with open(filename) as datafile:
                    hsr_df = read_by_line(datafile, hsr_date)
            except:
                return []
    
    return hsr_df[hsr_df.index.notnull()]

def load_Raw_series(hsr_dates, hsr_path, Series_filename, Compress = False):
    # hsr_dates is a list of dates to load
    # hsr_path is the root of the datastructure
    # Series_filename is the name of the HDF filename to save the data to, in hsr_path
    #load up raw values - do a day at a time
    m_Raw = []
    Raw = []
    for i in range(0,8):
        m_Raw.append([])
        Raw.append([])
    for hsr_date in hsr_dates:
        for i in range(0,8):
            filename = 'Raw {}.txt'.format(i)
            fRaw = open_hsr_file(hsr_path, hsr_date, filename, True)
            if len(fRaw):
                if Compress == True:    #reduce to 1-min averages
                    fRaw = fRaw.astype(float).resample('1T',label = 'right', closed = 'right').mean().dropna(how='all')
                m_Raw[i].append(fRaw)
        print("loaded Raw files for " + hsr_date)
    for i in range(0,8):
        Raw[i] = pd.concat(m_Raw[i])
        keyname = 'Raw{}'.format(i)
        Raw[i].to_hdf(Series_filename, key = keyname)
    return Raw

def time_WL_plot(spectra_1,spectra_2,name_1,name_2,timestamp,wl_range): # function for 2D colorplots
    # inputs: spectra_1 and spectra_2 are 2D np.arrays (rows==timestamp, columns==wavelength),
    # name 1 and 2 are corresponding variable names, and timestamp is a datetimelist
    
    datestamp=timestamp[0].date() # date
    spec_D=(spectra_1-spectra_2).T   # tranposed spectral difference
    
    pl.figure(figsize=(18,14))

    #  timestamp plot limits - 6 and 16 UTC are default
    llim=mdates.date2num(dt.datetime(datestamp.year,datestamp.month,datestamp.day,4,0,0))  
    ulim=mdates.date2num(dt.datetime(datestamp.year,datestamp.month,datestamp.day,20,0,0))
    
    time_day_ar = np.array(timestamp) # convert to np array format
    wl_ar = np.array(wl_range)
    
    pl.title(name_1 +' - ' + name_2 + ':  {}'.format(datestamp.strftime('%Y-%m-%d')),size=18)
    pl.pcolormesh(time_day_ar,wl_ar,spec_D,cmap='bwr')
    #pl.pcolormesh(time_day_ar,wl_range,spec_D,cmap='bwr') $ no limits
    pl.gca().xaxis.set_major_formatter(mdates.DateFormatter('%H'))
    pl.gca().yaxis.set_major_locator(pl.MultipleLocator(100) )
    pl.tight_layout()
    pl.gca().invert_yaxis()
    pl.ylabel("Wavelength [nm]")
    pl.xlabel("UTC time [hrs]")
    pl.xlim(llim,ulim)
    #pl.ylim(350,900)
    #pl.yaxis.set_major_locator(ticker.AutoLocator())
    cbar=pl.colorbar(orientation="horizontal")
    cbar.set_label(name_1 +' - ' + name_2 + ' [$mW m^{-2} nm^{-1}]$',size=18)



def time_WL_plot2(dfSpectrum, title): # function for 2D colorplots
    # inputs: dfSpectrum is a Pandas dataframe containing time v spectral data
    # title is some text to go in the title
    
    wl_range=dfSpectrum.columns
    timestamp = dfSpectrum.index   

    datestamp=timestamp[0].date() # date
    spec_D=dfSpectrum.T   # tranposed spectrum
    
    pl.figure(figsize=(18,14))

    #  timestamp plot limits - 6 and 16 UTC are default
    llim=mdates.date2num(dt.datetime(datestamp.year,datestamp.month,datestamp.day,15,0,0))  
    ulim=mdates.date2num(dt.datetime(datestamp.year,datestamp.month,datestamp.day,17,0,0))
    
    time_day_ar = np.array(timestamp) # convert to np array format
    wl_ar = np.array(wl_range)
    
    pl.title(title + ':  {}'.format(datestamp.strftime('%Y-%m-%d')),size=18)
    pl.pcolormesh(time_day_ar,wl_ar,spec_D,cmap='bwr')
    #pl.pcolormesh(time_day_ar,wl_range,spec_D,cmap='bwr') $ no limits
    pl.gca().xaxis.set_major_formatter(mdates.DateFormatter('%H'))
    pl.gca().yaxis.set_major_locator(pl.MultipleLocator(100) )
    pl.tight_layout()
    pl.gca().invert_yaxis()
    pl.ylabel("Wavelength [nm]")
    pl.xlabel("UTC time [hrs]")
    pl.xlim(llim,ulim)
    #pl.ylim(350,900)
    #pl.yaxis.set_major_locator(ticker.AutoLocator())
    cbar=pl.colorbar(orientation="horizontal")
    cbar.set_label('Intensity' + ' [$mW m^{-2} nm^{-1}]$',size=18)





if __name__ == '__main__':
    
    """    
     hsr_path='D:\DOCUMENT\BF4 Spectral\SolarSpectrumLogging\HSR1-003\Danube 2021' # path to directory containing all the hsr .txt files
     hsr_date='2021-10-11'    # clean files
     
    # clean = open_gps_file(hsr_path, hsr_date)
     total = open_hsr_file(hsr_path, hsr_date, 'Total.txt')
     diffuse = open_hsr_file(hsr_path, hsr_date, 'Diffuse.txt')
     summary = open_hsr_file(hsr_path, hsr_date, 'Summary.txt')
    # gps = open_hsr_file(hsr_path, hsr_date, 'GPS.txt')
     hsr_date='2020-10-29'    # broken files
    # broken = open_gps_file(hsr_path, hsr_date)
    """

    hsr_path='D:\DOCUMENT\BF4 Spectral\SolarSpectrumLogging\HSR1-002\PMLQuest 2021' # path to directory containing all the hsr .txt files
    hsr_date='2021-08-03'    # 
    lat = 50.37
    lon = -4.13
    Ed = open_hsr_file(hsr_path, hsr_date, 'Total.txt')
    #Eds = open_hsr_file(hsr_path, hsr_date, 'Diffuse.txt')
    Summary = open_hsr_file(hsr_path, hsr_date, 'Summary.txt')
    
    time_WL_plot2(Ed, 'Ed only')
    """
    hsr_path='D:\DOCUMENT\BF4 Spectral\SolarSpectrumLogging\HSR1-001\Winster 2021' # path to directory containing all the hsr .txt files
    hsr_date='2021-09-21'
    filename = 'Total.txt'
    Ed = open_hsr_file(hsr_path, hsr_date, 'Total.txt')
    Eds = open_hsr_file(hsr_path, hsr_date, 'Diffuse.txt')
    wl_range=Ed.columns
    timestamp = Ed.index   
    time_WL_plot(Ed, Eds, 'Ed', 'Eds',timestamp, wl_range )
    time_WL_plot2(Ed, 'Ed only')
    """



